#  Copyright (c) 2023 Roboto Technologies, Inc.
from ..command import RobotoCommandSet
from .logs import get_logs_command
from .show import show_command
from .status import status_command

commands = [
    get_logs_command,
    show_command,
    status_command,
]

command_set = RobotoCommandSet(
    name="invocations",
    help=("Show details of action invocations, their status history, and logs."),
    commands=commands,
)



import abc
from typing import Optional

from ...http import PaginatedList
from .record import CommentRecord, EntityType


class CommentDelegate(abc.ABC):
    @abc.abstractmethod
    def create_comment(
        self,
        entity_type: EntityType,
        entity_id: str,
        comment_text: str,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,
    ) -> CommentRecord:
        raise NotImplementedError("create_comment")

    @abc.abstractmethod
    def get_comment_by_id(
        self,
        comment_id: str,
        org_id: Optional[str] = None,
    ) -> CommentRecord:
        raise NotImplementedError("get_comment_by_id")

    @abc.abstractmethod
    def delete_comment(
        self,
        record: CommentRecord,
    ) -> None:
        raise NotImplementedError("delete_comment")

    @abc.abstractmethod
    def update_comment(
        self,
        record: CommentRecord,
        comment_text: str,
    ) -> CommentRecord:
        raise NotImplementedError("update_comment")

    @abc.abstractmethod
    def get_comments_by_entity(
        self,
        entity_type: EntityType,
        entity_id: str,
        org_id: Optional[str] = None,
        page_token: Optional[str] = None,
    ) -> PaginatedList[CommentRecord]:
        raise NotImplementedError("get_comments_by_entity")



from .conditions import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
    ConditionType,
)
from .query import (
    QuerySpecification,
    SortDirection,
)
from .visitor import BaseVisitor, ConditionVisitor

__all__ = (
    "BaseVisitor",
    "Comparator",
    "Condition",
    "ConditionGroup",
    "ConditionOperator",
    "ConditionType",
    "ConditionVisitor",
    "QuerySpecification",
    "SortDirection",
)



#  Copyright (c) 2023 Roboto Technologies, Inc.
import enum
import typing
from typing import Any, Optional, Union

import pydantic

from roboto.sentinels import (
    NotSet,
    NotSetType,
)

from ...pydantic import (
    validate_nonzero_gitpath_specs,
)
from ...query import ConditionType
from ..actions import (
    ComputeRequirements,
    ContainerParameters,
)
from .trigger_record import (
    TriggerForEachPrimitive,
)


class CreateTriggerRequest(pydantic.BaseModel):
    # Required
    name: str = pydantic.Field(regex=r"[\w\-]{1,256}")
    action_name: str
    action_owner_id: Optional[str] = None
    action_digest: Optional[str] = None
    required_inputs: list[str]
    for_each: TriggerForEachPrimitive
    compute_requirement_overrides: Optional[ComputeRequirements] = None
    container_parameter_overrides: Optional[ContainerParameters] = None
    condition: Optional[ConditionType] = None
    additional_inputs: Optional[list[str]] = None
    parameter_values: Optional[dict[str, Any]] = None

    @pydantic.validator("required_inputs")
    def validate_required_inputs(cls, value: list[str]) -> list[str]:
        return validate_nonzero_gitpath_specs(value)

    @pydantic.validator("additional_inputs")
    def validate_additional_inputs(
        cls, value: Optional[list[str]]
    ) -> Optional[list[str]]:
        return None if value is None else validate_nonzero_gitpath_specs(value)


class QueryTriggersRequest(pydantic.BaseModel):
    filters: dict[str, Any] = pydantic.Field(default_factory=dict)

    class Config:
        extra = "forbid"


class UpdateTriggerRequest(pydantic.BaseModel):
    action_name: Union[str, NotSetType] = NotSet
    action_owner_id: Union[str, NotSetType] = NotSet
    action_digest: Optional[Union[str, NotSetType]] = NotSet
    required_inputs: Union[list[str], NotSetType] = NotSet
    for_each: Union[TriggerForEachPrimitive, NotSetType] = NotSet
    enabled: Union[bool, NotSetType] = NotSet
    additional_inputs: Optional[Union[list[str], NotSetType]] = NotSet
    compute_requirement_overrides: Optional[
        Union[ComputeRequirements, NotSetType]
    ] = NotSet
    container_parameter_overrides: Optional[
        Union[ContainerParameters, NotSetType]
    ] = NotSet
    condition: Optional[Union[ConditionType, NotSetType]] = NotSet
    parameter_values: Optional[Union[dict[str, Any], NotSetType]] = NotSet

    class Config:
        extra = "forbid"
        schema_extra = NotSetType.openapi_schema_modifier


class EvaluateTriggerPrincipalType(str, enum.Enum):
    Dataset = "dataset"
    File = "file"


class EvaluateTriggerScope(str, enum.Enum):
    Dataset = "dataset"
    DatasetFiles = "dataset_files"
    File = "file"


class EvaluateTriggersRequest(pydantic.BaseModel):
    principal_id: str
    principal_type: EvaluateTriggerPrincipalType
    evaluation_scope: EvaluateTriggerScope

    @staticmethod
    def is_valid_combination(
        principal_type: EvaluateTriggerPrincipalType,
        evaluation_scope: EvaluateTriggerScope,
    ) -> bool:
        return (principal_type, evaluation_scope) in [
            (EvaluateTriggerPrincipalType.File, EvaluateTriggerScope.File),
            (EvaluateTriggerPrincipalType.Dataset, EvaluateTriggerScope.Dataset),
            (EvaluateTriggerPrincipalType.Dataset, EvaluateTriggerScope.DatasetFiles),
        ]

    @pydantic.validator("evaluation_scope")
    def validate_evaluation_scope(
        cls, evaluation_scope: EvaluateTriggerScope, values: dict[str, Any]
    ) -> EvaluateTriggerScope:
        principal_type = typing.cast(
            EvaluateTriggerPrincipalType, values.get("principal_type")
        )
        if not cls.is_valid_combination(principal_type, evaluation_scope):
            raise ValueError(
                f"'{principal_type}', '{evaluation_scope}' is not a valid tuple of "
                + "principal_type, evaluation_scope"
            )

        return evaluation_scope



#  Copyright (c) 2023 Roboto Technologies, Inc.
import enum
import json
import os
import pathlib
from typing import Any, Optional

import pydantic

from ..logging import default_logger

log = default_logger()

PROFILE_ENV_VAR = "ROBOTO_PROFILE"

PROD_USER_POOL_CLIENT_ID = "1gricmdmh0vv582qdd84phab5"
PROD_ENDPOINT = "https://api.roboto.ai"

DEFAULT_ROBOTO_CONFIG_DIR = pathlib.Path.home() / ".roboto"
DEFAULT_ROBOTO_PROFILE_FILE = DEFAULT_ROBOTO_CONFIG_DIR / "config.json"


class RobotoProfileEntry(pydantic.BaseModel):
    user_id: str
    token: str
    default_endpoint: str = PROD_ENDPOINT
    default_client_id: str = PROD_USER_POOL_CLIENT_ID


class RobotoProfileFileType(str, enum.Enum):
    none = "none"
    malformed = "malformed"
    implicit = "implicit"
    explicit = "explicit"


class MissingProfileException(BaseException):
    pass


class MalformedProfileException(BaseException):
    pass


class RobotoProfile:
    __config_file: pathlib.Path
    __default_profile_name: str

    def __init__(
        self,
        config_file: pathlib.Path = DEFAULT_ROBOTO_PROFILE_FILE,
        default_profile_name: Optional[str] = None,
    ):
        self.__config_file = config_file
        self.__default_profile_name = (
            default_profile_name
            if default_profile_name is not None
            else os.environ.get(PROFILE_ENV_VAR, "default")
        )

    @property
    def config_dir(self) -> pathlib.Path:
        return self.__config_file.parent

    @property
    def config_file(self) -> pathlib.Path:
        return self.__config_file

    def get_entry(self, profile_name: Optional[str] = None) -> RobotoProfileEntry:
        profile_to_check = (
            profile_name if profile_name is not None else self.__default_profile_name
        )
        log.debug("Using profile '%s'", profile_to_check)
        # TODO - Support AWS style order of precedence
        file_type, entry = self.__base_entry_from_file(profile_to_check)

        if file_type is RobotoProfileFileType.malformed:
            raise MalformedProfileException(
                f"Malformed roboto profile file '{self.__config_file}'"
            )
        elif file_type is RobotoProfileFileType.none:
            raise MissingProfileException(
                f"Missing roboto profile file '{self.__config_file}'"
            )

        assert entry is not None
        return entry

    def __base_entry_from_file(
        self,
        profile_name: str,
    ) -> tuple[RobotoProfileFileType, Optional[RobotoProfileEntry]]:
        if not self.__config_file.is_file():
            return RobotoProfileFileType.none, None

        with open(self.__config_file, "r") as f:
            try:
                contents: dict[str, Any] = json.loads(f.read())
            except json.decoder.JSONDecodeError:
                return RobotoProfileFileType.malformed, None

        if profile_name in contents.keys():
            try:
                return RobotoProfileFileType.explicit, RobotoProfileEntry.parse_obj(
                    contents.get(profile_name)
                )
            except pydantic.ValidationError:
                log.warning(
                    f"Couldn't parse {self.__config_file} as a multi-profile 'explicit' type"
                )
                return RobotoProfileFileType.malformed, None

        try:
            return RobotoProfileFileType.implicit, RobotoProfileEntry.parse_obj(
                contents
            )
        except pydantic.ValidationError:
            log.warning(
                f"Couldn't parse {self.__config_file} as single-profile 'implicit' type"
            )
            return RobotoProfileFileType.malformed, None



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...serde import pydantic_jsonable_dict
from .user_delegate import UserDelegate
from .user_record import UserRecord


class User:
    __record: UserRecord
    __user_delegate: UserDelegate

    @classmethod
    def from_id(cls, user_delegate: UserDelegate, user_id: Optional[str] = None):
        record = user_delegate.get_user_by_id(user_id=user_id)
        return cls(record=record, user_delegate=user_delegate)

    def __init__(self, record: UserRecord, user_delegate: UserDelegate):
        self.__record = record
        self.__user_delegate = user_delegate

    def to_dict(self, exclude_none: bool = False) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record, exclude_none=exclude_none)

    def delete(self) -> None:
        return self.__user_delegate.delete_user(user_id=self.__record.user_id)

    def update(self, updates: dict[str, Any]) -> UserRecord:
        self.__record = self.__user_delegate.update_user(
            user_id=self.__record.user_id, updates=updates
        )
        return self.__record

    @property
    def user_id(self) -> str:
        return self.__record.user_id

    @property
    def username(self) -> str:
        return self.__record.user_id

    @property
    def is_system_user(self) -> bool:
        return (
            self.__record.is_system_user
            if self.__record.is_system_user is not None
            else False
        )

    @property
    def name(self) -> Optional[str]:
        return self.__record.name

    @property
    def picture_url(self) -> Optional[str]:
        return self.__record.picture_url

    @property
    def record(self) -> UserRecord:
        return self.__record



#  Copyright (c) 2023 Roboto Technologies, Inc.

import dataclasses

from ..http import HttpClient, PATAuthDecorator
from ..profile import RobotoProfile
from .actions import (
    ActionHttpDelegate,
    InvocationHttpDelegate,
)
from .comments import CommentHttpDelegate
from .datasets import DatasetHttpDelegate
from .files import FileClientDelegate
from .orgs import OrgHttpDelegate
from .tokens import TokenHttpDelegate
from .triggers import TriggerHttpDelegate
from .users import UserHttpDelegate


@dataclasses.dataclass(frozen=True)
class HttpDelegates:
    http_client: HttpClient

    actions: ActionHttpDelegate
    datasets: DatasetHttpDelegate
    files: FileClientDelegate
    invocations: InvocationHttpDelegate
    orgs: OrgHttpDelegate
    tokens: TokenHttpDelegate
    triggers: TriggerHttpDelegate
    users: UserHttpDelegate
    comments: CommentHttpDelegate

    @staticmethod
    def from_client(http: HttpClient, endpoint: str) -> "HttpDelegates":
        # Take endpoint explicitly
        actions = ActionHttpDelegate(roboto_service_base_url=endpoint, http_client=http)
        datasets = DatasetHttpDelegate(
            roboto_service_base_url=endpoint, http_client=http
        )
        files = FileClientDelegate(roboto_service_base_url=endpoint, http_client=http)
        invocations = InvocationHttpDelegate(
            roboto_service_base_url=endpoint, http_client=http
        )

        # Take endpoint implicitly through client
        orgs = OrgHttpDelegate(http_client=http)
        tokens = TokenHttpDelegate(http_client=http)
        triggers = TriggerHttpDelegate(http_client=http)
        users = UserHttpDelegate(http_client=http)

        comments = CommentHttpDelegate(http_client=http)

        return HttpDelegates(
            http_client=http,
            actions=actions,
            datasets=datasets,
            files=files,
            invocations=invocations,
            orgs=orgs,
            tokens=tokens,
            triggers=triggers,
            users=users,
            comments=comments,
        )

    @staticmethod
    def from_profile(profile: RobotoProfile, entry_name: str = "default"):
        entry = profile.get_entry(entry_name)
        auth = PATAuthDecorator.for_client(
            client_id=entry.default_client_id, profile=profile
        )
        http = HttpClient(default_endpoint=entry.default_endpoint, default_auth=auth)
        return HttpDelegates.from_client(http=http, endpoint=entry.default_endpoint)



import datetime
import enum
import json
from typing import Any, Optional

import pydantic

from ...serde import pydantic_jsonable_dict
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)


class InvocationDataSourceType(enum.Enum):
    """Source of data for an Action's InputBinding"""

    Dataset = "Dataset"


class InvocationDataSource(pydantic.BaseModel):
    data_source_type: InvocationDataSourceType
    # The "type" determines the meaning of "id":
    #   - if type is "Dataset," id is a dataset_id
    data_source_id: str


class ActionProvenance(pydantic.BaseModel):
    name: str
    org_id: str
    # 2023-09-11 (GM): Optional for backwards compatibility; new invocations will always have a digest
    digest: Optional[str] = None


class ExecutableProvenance(pydantic.BaseModel):
    container_image_uri: Optional[str] = None  # Optional for backwards compatibility
    container_image_digest: Optional[str] = None


class InvocationSource(enum.Enum):
    Trigger = "Trigger"
    Manual = "Manual"


class SourceProvenance(pydantic.BaseModel):
    source_type: InvocationSource
    # The “type” determines the meaning of “id:”
    #   - if type is “Trigger,” id is a TriggerId;
    #   - if type is “Manual,” id is a UserId.
    source_id: str


class InvocationProvenance(pydantic.BaseModel):
    action: ActionProvenance
    """The Action that was invoked."""

    executable: ExecutableProvenance
    """The underlying executable (e.g., Docker image) that was run."""

    source: SourceProvenance
    """The source of the invocation."""


class InvocationStatus(enum.Enum):
    Queued = 0
    Scheduled = 1
    Downloading = 2
    Processing = 3
    Uploading = 4
    Completed = 5
    # Failure status' exist outside linear progression of invocation status
    Failed = 998
    Deadly = 999

    def __str__(self) -> str:
        return self.name

    def can_transition_to(self, other: "InvocationStatus") -> bool:
        if self == other:
            return True

        if self in {InvocationStatus.Completed, InvocationStatus.Deadly}:
            return False

        if self is InvocationStatus.Failed:
            if other in {InvocationStatus.Queued, InvocationStatus.Deadly}:
                return True
            return False

        if other is InvocationStatus.Failed:
            return True

        if other is InvocationStatus.Deadly:
            return self in {InvocationStatus.Queued, InvocationStatus.Failed}

        return other.value - self.value == 1

    def is_running(self) -> bool:
        return self in {
            InvocationStatus.Downloading,
            InvocationStatus.Processing,
            InvocationStatus.Uploading,
        }

    def is_terminal(self) -> bool:
        return self in {
            InvocationStatus.Completed,
            InvocationStatus.Failed,
            InvocationStatus.Deadly,
        }

    def next(self) -> Optional["InvocationStatus"]:
        if self.is_terminal():
            return None
        return InvocationStatus(self.value + 1)


class InvocationStatusRecord(pydantic.BaseModel):
    status: InvocationStatus
    detail: Optional[str] = None
    timestamp: datetime.datetime  # Persisted as ISO 8601 string in UTC

    def to_presentable_dict(self) -> dict[str, Optional[str]]:
        return {
            "status": str(self.status),
            "timestamp": self.timestamp.isoformat(),
            "detail": self.detail,
        }


class LogsLocation(pydantic.BaseModel):
    bucket: str
    prefix: str


class InvocationRecord(pydantic.BaseModel):
    # When adding or removing fields, make sure to update __str__
    created: datetime.datetime  # Persisted as ISO 8601 string in UTC
    data_source: InvocationDataSource
    input_data: list[str]
    invocation_id: str  # Sort key
    idempotency_id: Optional[str] = None
    compute_requirements: ComputeRequirements
    container_parameters: ContainerParameters
    logs: Optional[LogsLocation] = None
    org_id: str  # Partition key
    parameter_values: dict[str, Any] = pydantic.Field(default_factory=dict)
    provenance: InvocationProvenance
    status: list[InvocationStatusRecord] = pydantic.Field(default_factory=list)

    def __str__(self) -> str:
        return json.dumps(
            {
                "created": self.created.isoformat(),
                "data_source": pydantic_jsonable_dict(self.data_source),
                "input_data": self.input_data,
                "invocation_id": self.invocation_id,
                "idempotency_id": self.idempotency_id,
                "logs": self.logs.dict() if self.logs else None,
                "compute_requirements": pydantic_jsonable_dict(
                    self.compute_requirements
                ),
                "container_parameters": pydantic_jsonable_dict(
                    self.container_parameters
                ),
                "org_id": self.org_id,
                "parameter_values": self.parameter_values,
                "provenance": pydantic_jsonable_dict(self.provenance),
                "status": [
                    status_record.to_presentable_dict() for status_record in self.status
                ],
            },
            indent=2,
        )


class LogRecord(pydantic.BaseModel):
    log: str
    timestamp: datetime.datetime



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Optional

from ...exceptions import RobotoHttpExceptionParse
from ...http import HttpClient, roboto_headers
from ...serde import pydantic_jsonable_dict
from .delegate import TokenDelegate
from .http_resources import CreateTokenRequest
from .record import TokenRecord


class TokenHttpDelegate(TokenDelegate):
    __http_client: HttpClient

    def __init__(self, http_client: HttpClient):
        super().__init__()
        self.__http_client = http_client

    def get_tokens_for_user(self, user_id: Optional[str] = None) -> list[TokenRecord]:
        url = self.__http_client.url("v1/tokens")
        headers = roboto_headers(user_id=user_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)

        unmarshalled = response.from_json(json_path=["data"])
        return [TokenRecord.parse_obj(record) for record in unmarshalled]

    def create_token(
        self,
        expiry_days: int,
        name: str,
        user_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> TokenRecord:
        url = self.__http_client.url("v1/tokens")
        headers = roboto_headers(
            user_id=user_id, additional_headers={"Content-Type": "application/json"}
        )

        data = CreateTokenRequest(
            expiry_days=expiry_days, name=name, description=description
        )

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url=url, headers=headers, data=pydantic_jsonable_dict(data)
            )
        return TokenRecord.parse_obj(response.from_json(json_path=["data"]))

    def delete_token(self, token_id: str) -> None:
        url = self.__http_client.url(f"v1/tokens/{token_id}")

        with RobotoHttpExceptionParse():
            self.__http_client.delete(url=url)

    def get_token_by_token_id(self, token_id: str) -> TokenRecord:
        url = self.__http_client.url(f"v1/tokens/{token_id}")

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)
        return TokenRecord.parse_obj(response.from_json(json_path=["data"]))



#  Copyright (c) 2023 Roboto Technologies, Inc.


import datetime
import pathlib
from typing import Any, Optional

import boto3
import jwt

from ..logging import default_logger
from ..profile import RobotoProfile

log = default_logger()

bearer_token_vendor = None
cred_finder = None

CREDENTIALS_FILE = pathlib.Path(f"{pathlib.Path.home()}/.roboto/credentials.json")
ROBOTO_TEMPDIR = pathlib.Path(f"{pathlib.Path.home()}/.roboto/tmp")
PROFILE_ENV_VAR = "ROBOTO_PROFILE"
USER_ENV_VAR = "ROBOTO_USER_ID"
TOKEN_ENV_VAR = "ROBOTO_ACCESS_TOKEN"


class RejectedTokenException(Exception):
    pass


class BearerTokenVendor:
    _client_id: str
    _cognito: Any
    _expires: datetime.datetime
    _id_token: Optional[str]
    _profile: RobotoProfile

    def __init__(
        self,
        client_id: str,
        cognito: Optional[Any] = None,
        profile: RobotoProfile = RobotoProfile(),
    ):
        self._client_id = client_id
        self._cognito = (
            cognito
            if cognito is not None
            else boto3.client("cognito-idp", region_name="us-west-2")
        )
        self._expires = datetime.datetime.now()
        self._id_token = None
        self._profile = profile

    def perform_cognito_handshake(self, user_id: str, token: str):
        try:
            res = self._cognito.initiate_auth(
                AuthFlow="CUSTOM_AUTH",
                ClientId=self._client_id,
                AuthParameters={"USERNAME": user_id},
            )

            if res["ChallengeName"] != "CUSTOM_CHALLENGE":
                raise RejectedTokenException(
                    f"Unexpected challenge {res['ChallengeName']}"
                )

            res = self._cognito.respond_to_auth_challenge(
                ClientId=self._client_id,
                ChallengeName="CUSTOM_CHALLENGE",
                Session=res["Session"],
                ChallengeResponses={"USERNAME": user_id, "ANSWER": token},
            )

            # Assume we have 60 fewer seconds to refresh than AWS is telling use we have
            expires_in = res["AuthenticationResult"]["ExpiresIn"] - 60
            expiry = datetime.datetime.now() + datetime.timedelta(seconds=expires_in)
            id_token = res["AuthenticationResult"]["IdToken"]

            return id_token, expiry
        except self._cognito.exceptions.NotAuthorizedException:
            raise RejectedTokenException("Authorization rejected by Cognito!")
        except self._cognito.exceptions.UserNotFoundException:
            raise RejectedTokenException(f"User {user_id} does not exist!")

    def refresh_tokens(self):
        now = datetime.datetime.now()
        entry = self._profile.get_entry()
        id_token_file = pathlib.Path(f"{ROBOTO_TEMPDIR}/{entry.token}_id_token")

        if id_token_file.exists():
            with open(id_token_file, "r") as f:
                content = f.read()

            # We don't need to verify signature because we'll get rejected by Roboto Service if the signature is invalid
            id_token = jwt.decode(content, options={"verify_signature": False})

            expires = datetime.datetime.fromtimestamp(id_token["exp"])
            if expires > now:
                self._expires = expires
                self._id_token = content
                return

        (
            self._id_token,
            self._expires,
        ) = self.perform_cognito_handshake(entry.user_id, entry.token)

        assert self._id_token is not None

        ROBOTO_TEMPDIR.mkdir(parents=True, exist_ok=True)
        with open(id_token_file, "w") as f:
            f.write(self._id_token)

    def get_id_token(self):
        now = datetime.datetime.now()
        if self._id_token is None or self._expires < now:
            log.debug("ID Token is unavailable or stale, regenerating it")
            self.refresh_tokens()
        else:
            log.debug("ID Token is already available!")

        return self._id_token

    def get_auth_header(self):
        return f"Bearer {self.get_id_token()}"



#  Copyright (c) 2023 Roboto Technologies, Inc.

import abc
from typing import Any, Optional

from .record import (
    OrgInviteRecord,
    OrgRecord,
    OrgRoleName,
    OrgRoleRecord,
)


class OrgDelegate(abc.ABC):
    @abc.abstractmethod
    def create_org(
        self,
        creator_user_id: Optional[str],
        name: str,
        bind_email_domain: bool = False,
    ) -> OrgRecord:
        raise NotImplementedError("create_org")

    @abc.abstractmethod
    def update_org(
        self,
        updates: dict[str, Any],
        org_id: Optional[str] = None,
        caller_user_id: Optional[str] = None,
    ) -> OrgRecord:
        raise NotImplementedError("update_org")

    @abc.abstractmethod
    def orgs_for_user(self, user_id: Optional[str]) -> list[OrgRecord]:
        raise NotImplementedError("orgs_for_user")

    @abc.abstractmethod
    def org_roles_for_user(self, user_id: Optional[str]) -> list[OrgRoleRecord]:
        raise NotImplementedError("org_roles_for_user")

    @abc.abstractmethod
    def org_roles_for_org(self, org_id: Optional[str]) -> list[OrgRoleRecord]:
        raise NotImplementedError("org_roles_for_org")

    @abc.abstractmethod
    def org_role_for_user_in_org(
        self, user_id: Optional[str] = None, org_id: Optional[str] = None
    ) -> OrgRoleRecord:
        raise NotImplementedError("org_role_for_user_in_org")

    @abc.abstractmethod
    def add_role_for_user(
        self, user_id: str, role_name: OrgRoleName, org_id: Optional[str] = None
    ):
        raise NotImplementedError("add_role_for_user")

    @abc.abstractmethod
    def remove_role_from_user(
        self, user_id: str, role_name: OrgRoleName, org_id: Optional[str] = None
    ):
        raise NotImplementedError("remove_role_for_user")

    @abc.abstractmethod
    def remove_user_from_org(self, user_id: str, org_id: Optional[str] = None) -> None:
        raise NotImplementedError("remove_user_from_org")

    @abc.abstractmethod
    def get_org_by_id(self, org_id: str) -> OrgRecord:
        raise NotImplementedError("get_org_by_id")

    @abc.abstractmethod
    def delete_org(self, org_id: str):
        raise NotImplementedError("delete_org")

    @abc.abstractmethod
    def bind_email_domain(self, org_id: str, email_domain: str):
        raise NotImplementedError("bind_email_domain")

    @abc.abstractmethod
    def unbind_email_domain(self, email_domain: str, org_id: Optional[str] = None):
        raise NotImplementedError("unbind_email_domain")

    @abc.abstractmethod
    def get_email_domains_for_org(self, org_id: Optional[str] = None) -> list[str]:
        raise NotImplementedError("get_email_domains_for_org")

    @abc.abstractmethod
    def invite_user_to_org(
        self, invited_user_id: str, org_id: str, inviting_user_id: Optional[str] = None
    ) -> OrgInviteRecord:
        raise NotImplementedError("invite_user_to_org")

    @abc.abstractmethod
    def get_invites_for_org(
        self, org_id: Optional[str] = None
    ) -> list[OrgInviteRecord]:
        raise NotImplementedError("get_invites_for_org")

    @abc.abstractmethod
    def accept_org_invite(self, invite_id: str, user_id: Optional[str] = None):
        raise NotImplementedError("accept_org_invite")

    @abc.abstractmethod
    def decline_org_invite(self, invite_id: str, user_id: Optional[str] = None):
        raise NotImplementedError("accept_org_invite")

    @abc.abstractmethod
    def get_org_invite(
        self, invite_id: str, user_id: Optional[str] = None
    ) -> OrgInviteRecord:
        raise NotImplementedError("get_org_invite")



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import json
import sys

from ...domain.actions import Action
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..common_args import (
    ActionParameterArg,
    ParseError,
    add_action_reference_arg,
    add_compute_requirements_args,
    add_container_parameters_args,
    add_org_arg,
    parse_compute_requirements,
    parse_container_overrides,
)
from ..context import CLIContext


def create(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    try:
        compute_requirements = parse_compute_requirements(args)
        container_parameters = parse_container_overrides(args)
    except ParseError as exc:
        print(exc.msg, file=sys.stderr)
    else:
        action = Action.create(
            name=args.name,
            parameters=args.parameter,
            uri=args.image,
            inherits=args.inherits_from,
            description=args.description,
            compute_requirements=compute_requirements,
            container_parameters=container_parameters,
            metadata=args.metadata,
            tags=args.tag,
            action_delegate=context.actions,
            invocation_delegate=context.invocations,
            org_id=args.org,
        )

        print(f"Successfully created action '{action.name}'. Record: ")
        print(json.dumps(action.to_dict(), indent=4))


def create_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--name",
        required=True,
        action="store",
        help=(
            "Name of the action. Not modifiable after creation. "
            "An action is considered unique by its (name, docker_image_name, docker_image_tag) tuple."
        ),
    )
    parser.add_argument(
        "--description",
        required=False,
        action="store",
        help="Optional description of action. Modifiable after creation.",
    )
    add_action_reference_arg(
        parser=parser,
        arg_name="inherits_from",
        arg_help=(
            "Partially or fully qualified reference to action from which to inherit configuration. "
            "Inheriting from another action is mutually exclusive with specifying a container image (--image), "
            "entrypoint (--entrypoint), command (--command), working directory (--workdir), env vars (--env), "
            "or parameter(s) (--parameter). "
        ),
        positional=False,
        required=False,
    )
    parser.add_argument(
        "--image",
        required=False,
        action="store",
        dest="image",
        help="Associate a Docker image with this action. Modifiable after creation.",
    )
    parser.add_argument(
        "--parameter",
        required=False,
        metavar=ActionParameterArg.METAVAR,
        nargs="*",
        action=ActionParameterArg,
        help=(
            "Zero or more parameters (space-separated) accepted by this action. "
            "'name' is the only required attribute. "
            "'default' values, if provided, are JSON parsed. "
            "This argument can be specified multiple times. "
            "Parameters can be modified after creation. "
            "Argument values must be wrapped in quotes. E.g.: "
            "--put-parameter 'name=my_param|required=true|description=My description of my_param'"
        ),
    )
    parser.add_argument(
        "--metadata",
        required=False,
        metavar="KEY=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more 'key=value' format key/value pairs which represent action metadata. "
            "`value` is parsed as JSON. "
            "Metadata can be modified after creation."
        ),
    )
    parser.add_argument(
        "--tag",
        required=False,
        type=str,
        nargs="*",
        help="One or more tags to annotate this action. Modifiable after creation.",
        action="extend",
    )
    add_org_arg(parser=parser)

    add_compute_requirements_args(parser)
    add_container_parameters_args(parser)


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_parser,
    command_kwargs={"help": "Create a new action."},
)



import abc
import datetime
from typing import Any, Optional

import pydantic

from ...auth import Permissions
from ...http import PaginatedList
from ...query import QuerySpecification
from ...serde import pydantic_jsonable_dict
from ...time import utcnow
from ...updates import (
    MetadataChangeset,
    UpdateCondition,
)
from ..files import FileRecord
from .record import (
    Administrator,
    DatasetRecord,
    StorageLocation,
)


class Credentials(pydantic.BaseModel):
    access_key_id: str
    bucket: str
    expiration: datetime.datetime
    secret_access_key: str
    session_token: str
    required_prefix: str
    upload_id: Optional[str] = None

    def is_expired(self) -> bool:
        return utcnow() >= self.expiration

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self, exclude_none=True)


class DatasetDelegate(abc.ABC):
    @abc.abstractmethod
    def complete_upload(
        self, dataset_id: str, upload_id: str, org_id: Optional[str] = None
    ) -> None:
        raise NotImplementedError("complete_upload")

    @abc.abstractmethod
    def create_dataset(
        self,
        administrator: Administrator = Administrator.Roboto,
        metadata: Optional[dict[str, Any]] = None,
        storage_location: StorageLocation = StorageLocation.S3,
        tags: Optional[list[str]] = None,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,
        description: Optional[str] = None,
    ) -> DatasetRecord:
        raise NotImplementedError("create_dataset")

    @abc.abstractmethod
    def delete_dataset(self, record: DatasetRecord) -> None:
        raise NotImplementedError("delete_dataset")

    @abc.abstractmethod
    def get_dataset_by_primary_key(
        self,
        dataset_id: str,
        org_id: Optional[str] = None,
    ) -> DatasetRecord:
        raise NotImplementedError("get_dataset_by_primary_key")

    @abc.abstractmethod
    def get_temporary_credentials(
        self,
        record: DatasetRecord,
        permissions: Permissions,
        caller: Optional[str] = None,
        transaction_id: Optional[str] = None,
    ) -> Credentials:
        raise NotImplementedError("get_temporary_credentials")

    @abc.abstractmethod
    def list_files(
        self,
        dataset_id: str,
        org_id: Optional[str] = None,
        page_token: Optional[str] = None,
    ) -> PaginatedList[FileRecord]:
        raise NotImplementedError("list_files")

    @abc.abstractmethod
    def query_datasets(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[DatasetRecord]:
        raise NotImplementedError("query_datasets")

    @abc.abstractmethod
    def update(
        self,
        record: DatasetRecord,
        metadata_changeset: Optional[MetadataChangeset] = None,
        conditions: Optional[list[UpdateCondition]] = None,
        description: Optional[str] = None,
        updated_by: Optional[str] = None,
    ) -> DatasetRecord:
        raise NotImplementedError("update")



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Optional

import pydantic


class UserRecord(pydantic.BaseModel):
    user_id: str
    is_system_user: Optional[bool] = False
    name: Optional[str] = None
    picture_url: Optional[str] = None



import pathlib
from typing import Any, Optional
import urllib.parse

import boto3
import botocore.config

from ...exceptions import RobotoHttpExceptionParse
from ...http import (
    HttpClient,
    PaginatedList,
    roboto_headers,
)
from ...query import QuerySpecification
from ...serde import pydantic_jsonable_dict
from .delegate import (
    FileDelegate,
    FileTag,
    S3Credentials,
)
from .progress import (
    NoopProgressMonitorFactory,
    ProgressMonitorFactory,
)
from .record import FileRecord


class FileClientDelegate(FileDelegate):
    __http_client: HttpClient
    __roboto_service_base_url: str

    @staticmethod
    def generate_s3_client(credentials: S3Credentials, tcp_keepalive: bool = True):
        session = boto3.Session(
            aws_access_key_id=credentials.access_key_id,
            aws_secret_access_key=credentials.secret_access_key,
            aws_session_token=credentials.session_token,
        )

        return session.client(
            "s3", config=botocore.config.Config(tcp_keepalive=tcp_keepalive)
        )

    def __init__(self, roboto_service_base_url: str, http_client: HttpClient) -> None:
        super().__init__()
        self.__http_client = http_client
        self.__roboto_service_base_url = roboto_service_base_url

    def delete_file(self, record: FileRecord) -> None:
        url = f"{self.__roboto_service_base_url}/v1/files/{record.file_id}"

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url,
                headers=roboto_headers(
                    org_id=record.org_id,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )

    def download_file(
        self,
        record: FileRecord,
        local_path: pathlib.Path,
        credentials: S3Credentials,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ) -> None:
        local_path.parent.mkdir(parents=True, exist_ok=True)
        s3_client = FileClientDelegate.generate_s3_client(credentials)

        res = s3_client.head_object(Bucket=record.bucket, Key=record.key)
        download_bytes = int(res.get("ContentLength", 0))

        progress_monitor = progress_monitor_factory.download_monitor(
            source=record.key, size=download_bytes
        )
        try:
            s3_client.download_file(
                Bucket=record.bucket,
                Key=record.key,
                Filename=str(local_path),
                Callback=progress_monitor.update,
            )
        finally:
            progress_monitor.close()

    def get_record_by_primary_key(
        self, file_id: str, org_id: Optional[str] = None
    ) -> FileRecord:
        url = f"{self.__roboto_service_base_url}/v1/files/record/{file_id}"

        with RobotoHttpExceptionParse():
            res = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id=org_id,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        return FileRecord.parse_obj(res.from_json(json_path=["data"]))

    def get_signed_url(self, record: FileRecord) -> str:
        url = f"{self.__roboto_service_base_url}/v1/files/{record.file_id}/signed-url"

        with RobotoHttpExceptionParse():
            res = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id=record.org_id,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        return res.from_json(json_path=["data", "url"])

    def query_files(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[FileRecord]:
        url = f"{self.__roboto_service_base_url}/v1/files/query"
        post_body = pydantic_jsonable_dict(query, exclude_none=True)
        with RobotoHttpExceptionParse():
            res = self.__http_client.post(
                url,
                data=post_body,
                headers=roboto_headers(
                    org_id=org_id,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )

        unmarshalled = res.from_json(json_path=["data"])
        return PaginatedList(
            items=[FileRecord.parse_obj(dataset) for dataset in unmarshalled["items"]],
            next_token=unmarshalled["next_token"],
        )

    def upload_file(
        self,
        local_path: pathlib.Path,
        bucket: str,
        key: str,
        credentials: S3Credentials,
        tags: Optional[dict[FileTag, str]] = None,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ) -> None:
        upload_file_args: dict[str, Any] = {
            "Filename": str(local_path),
            "Key": key,
            "Bucket": bucket,
        }

        if tags is not None:
            serializable_tags = {tag.value: value for tag, value in tags.items()}
            encoded_tags = urllib.parse.urlencode(serializable_tags)
            upload_file_args["ExtraArgs"] = {"Tagging": encoded_tags}

        progress_monitor = progress_monitor_factory.upload_monitor(
            source=key, size=local_path.stat().st_size
        )
        upload_file_args["Callback"] = progress_monitor.update

        try:
            s3_client = FileClientDelegate.generate_s3_client(credentials)
            s3_client.upload_file(**upload_file_args)
        finally:
            if progress_monitor is not None:
                progress_monitor.close()



from .client_delegate import FileClientDelegate
from .delegate import (
    FileDelegate,
    FileTag,
    S3Credentials,
)
from .file import File
from .http_resources import (
    DeleteFileRequest,
    FileRecordRequest,
    QueryFilesRequest,
    SignedUrlResponse,
)
from .progress import (
    NoopProgressMonitor,
    NoopProgressMonitorFactory,
    ProgressMonitor,
    ProgressMonitorFactory,
    TqdmProgressMonitor,
    TqdmProgressMonitorFactory,
)
from .record import FileRecord

__all__ = (
    "DeleteFileRequest",
    "File",
    "FileClientDelegate",
    "FileDelegate",
    "FileRecord",
    "FileRecordRequest",
    "FileTag",
    "NoopProgressMonitor",
    "NoopProgressMonitorFactory",
    "ProgressMonitor",
    "ProgressMonitorFactory",
    "QueryFilesRequest",
    "S3Credentials",
    "SignedUrlResponse",
    "TqdmProgressMonitor",
    "TqdmProgressMonitorFactory",
)



import collections.abc
import enum
import typing

import pydantic


class ExecutorContainer(enum.Enum):
    LogRouter = "firelens_log_router"
    Monitor = "monitor"
    Setup = "setup"
    Action = "action"
    OutputHandler = "output_handler"


class ComputeRequirements(pydantic.BaseModel):
    """
    Compute requirements for an action invocation.

    .. _Relevant AWS Fargate documentation:
        https://docs.aws.amazon.com/AmazonECS/latest/developerguide/AWS_Fargate.html#fargate-tasks-size
    """

    vCPU: int = 512  # 256, 512, 1024, 2048, 4096, 8192, 16384
    memory: int = 1024  # 512, 1024, 2024, ... (120 * 1024) in MiB
    gpu: typing.Literal[False] = False  # Not yet supported
    storage: int = pydantic.Field(default=21, ge=21, le=200)  # in GiB (min 21, max 200)

    @pydantic.root_validator
    def validate_vcpu_mem_combination(cls, values: dict) -> dict:
        v_cpu = values["vCPU"]
        allowed_vpcu = (256, 512, 1024, 2048, 4096, 8192, 16384)
        if v_cpu not in allowed_vpcu:
            raise ValueError(f"Unsupported vCPU value. Allowed options: {allowed_vpcu}")

        memory = values["memory"]
        allowed_memory: collections.abc.Sequence = list()
        if v_cpu == 256:
            allowed_memory = [512, 1024, 2048]
        elif v_cpu == 512:
            allowed_memory = range(1024, 5 * 1024, 1024)
        elif v_cpu == 1024:
            allowed_memory = range(2 * 1024, 9 * 1024, 1024)
        elif v_cpu == 2048:
            allowed_memory = range(4 * 1024, 17 * 1024, 1024)
        elif v_cpu == 4096:
            allowed_memory = range(8 * 1024, 31 * 1024, 1024)
        elif v_cpu == 8192:
            allowed_memory = range(16 * 1024, 61 * 1024, 4 * 1024)
        elif v_cpu == 16384:
            allowed_memory = range(32 * 1024, 121 * 1024, 8 * 1024)
        else:
            raise ValueError(f"Unknown vCPU value {v_cpu}")

        if memory not in allowed_memory:
            raise ValueError(
                f"Unsupported memory/vCPU combination, allowed memory for {v_cpu} vCPU: {list(allowed_memory)}"
            )

        return values


class ContainerParameters(pydantic.BaseModel):
    command: typing.Optional[list[str]] = None
    entry_point: typing.Optional[list[str]] = None
    env_vars: typing.Optional[dict[str, str]] = None
    workdir: typing.Optional[str] = None



import abc
from typing import Any, Optional

from ...http import PaginatedList, StreamedList
from ...query import QuerySpecification
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .action_record import ActionRecord
from .invocation_record import (
    InvocationDataSourceType,
    InvocationRecord,
    InvocationSource,
    InvocationStatus,
    LogRecord,
    LogsLocation,
)


class InvocationDelegate(abc.ABC):
    @abc.abstractmethod
    def create_invocation(
        self,
        action_record: ActionRecord,
        parameter_values: dict[str, Any],
        input_data: list[str],
        data_source_id: str,
        data_source_type: InvocationDataSourceType,
        invocation_source: InvocationSource,
        invocation_source_id: Optional[str] = None,
        compute_requirement_overrides: Optional[ComputeRequirements] = None,
        container_parameter_overrides: Optional[ContainerParameters] = None,
        idempotency_id: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> InvocationRecord:
        raise NotImplementedError("create_invocation")

    @abc.abstractmethod
    def get_by_id(
        self, invocation_id: str, org_id: Optional[str] = None
    ) -> InvocationRecord:
        raise NotImplementedError("get_by_id")

    @abc.abstractmethod
    def get_logs(
        self,
        invocation_id: str,
        page_token: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> PaginatedList[LogRecord]:
        raise NotImplementedError("get_logs")

    @abc.abstractmethod
    def stream_logs(
        self,
        invocation_id: str,
        last_read: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> StreamedList[LogRecord]:
        raise NotImplementedError("stream_logs")

    @abc.abstractmethod
    def set_logs_location(
        self,
        record: InvocationRecord,
        logs: LogsLocation,
    ) -> InvocationRecord:
        raise NotImplementedError("set_logs_location")

    @abc.abstractmethod
    def update_invocation_status(
        self,
        record: InvocationRecord,
        status: InvocationStatus,
        detail: Optional[str] = None,
    ) -> InvocationRecord:
        raise NotImplementedError("update_invocation_status")

    @abc.abstractmethod
    def query_invocations(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[InvocationRecord]:
        raise NotImplementedError("query_invocations")



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...serde import pydantic_jsonable_dict
from .delegate import OrgDelegate
from .org import Org
from .record import OrgInviteRecord


class OrgInvite:
    __org_delegate: OrgDelegate
    __org: Org
    __record: OrgInviteRecord

    @classmethod
    def create(
        cls,
        invited_user_id: str,
        org_id: str,
        org_delegate: OrgDelegate,
        inviting_user_id: Optional[str] = None,
    ) -> "OrgInvite":
        record = org_delegate.invite_user_to_org(
            invited_user_id=invited_user_id,
            inviting_user_id=inviting_user_id,
            org_id=org_id,
        )
        return cls(record=record, org_delegate=org_delegate)

    @classmethod
    def from_id(
        cls, invite_id: str, org_delegate: OrgDelegate, user_id: Optional[str] = None
    ):
        record = org_delegate.get_org_invite(invite_id=invite_id, user_id=user_id)
        return cls(record=record, org_delegate=org_delegate)

    @classmethod
    def for_org(
        cls, org_delegate: OrgDelegate, org_id: Optional[str] = None
    ) -> list["OrgInvite"]:
        records = org_delegate.get_invites_for_org(org_id=org_id)
        return [cls(record=record, org_delegate=org_delegate) for record in records]

    def accept(self, user_id: Optional[str]):
        self.__org_delegate.accept_org_invite(
            invite_id=self.__record.invite_id, user_id=user_id
        )

    def decline(self, user_id: Optional[str]):
        self.__org_delegate.decline_org_invite(
            invite_id=self.__record.invite_id, user_id=user_id
        )

    def __init__(self, record: OrgInviteRecord, org_delegate: OrgDelegate):
        self.__record = record
        self.__org_delegate = org_delegate
        self.__org = Org(record=record.org, org_delegate=org_delegate)

    @property
    def invite_id(self) -> str:
        return self.__record.invite_id

    @property
    def target_user_id(self) -> str:
        return self.__record.user_id

    @property
    def invited_by_user_id(self) -> str:
        return self.__record.invited_by.user_id

    @property
    def org_id(self) -> str:
        return self.__record.org.org_id

    @property
    def record(self) -> OrgInviteRecord:
        return self.__record

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)



from typing import Optional

from ...exceptions import RobotoHttpExceptionParse
from ...http import (
    HttpClient,
    PaginatedList,
    roboto_headers,
)
from ...serde import pydantic_jsonable_dict
from .delegate import CommentDelegate
from .http_resources import (
    CreateCommentRequest,
    UpdateCommentRequest,
)
from .record import CommentRecord, EntityType


class CommentHttpDelegate(CommentDelegate):
    __http_client: HttpClient

    def __init__(self, http_client: HttpClient):
        super().__init__()
        self.__http_client = http_client

    def create_comment(
        self,
        entity_type: EntityType,
        entity_id: str,
        comment_text: str,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,
    ) -> CommentRecord:
        url = self.__http_client.url("v1/comments")

        headers = roboto_headers(
            org_id=org_id,
            user_id=created_by,
            additional_headers={"Content-Type": "application/json"},
        )

        request_body = CreateCommentRequest(
            entity_type=entity_type,
            entity_id=entity_id,
            comment_text=comment_text,
        )

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

        return CommentRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_comment_by_id(
        self,
        comment_id: str,
        org_id: Optional[str] = None,
    ) -> CommentRecord:
        url = self.__http_client.url(f"v1/comments/{comment_id}")

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)

        return CommentRecord.parse_obj(response.from_json(json_path=["data"]))

    def delete_comment(
        self,
        record: CommentRecord,
    ) -> None:
        url = self.__http_client.url(f"v1/comments/{record.comment_id}")

        with RobotoHttpExceptionParse():
            self.__http_client.delete(url=url)

    def update_comment(
        self,
        record: CommentRecord,
        comment_text: str,
    ) -> CommentRecord:
        url = self.__http_client.url(f"v1/comments/{record.comment_id}")

        req = UpdateCommentRequest(
            comment_text=comment_text,
        )

        body = pydantic_jsonable_dict(req, exclude_unset=True)

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url=url,
                data=body,
            )

        return CommentRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_comments_by_entity(
        self,
        entity_type: EntityType,
        entity_id: str,
        org_id: Optional[str] = None,
        page_token: Optional[str] = None,
    ) -> PaginatedList[CommentRecord]:
        url = self.__http_client.url(f"v1/comments/{entity_type}/{entity_id}")

        if page_token:
            url = f"{url}?page_token={page_token}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)

        parsed_items = []

        unmarshalled = response.from_json(json_path=["data"])

        data_items = unmarshalled["items"]
        next_token = unmarshalled["next_token"]

        for item in data_items:
            parsed_items.append(CommentRecord.parse_obj(item))

        return PaginatedList(items=parsed_items, next_token=next_token)



#  Copyright (c) 2023 Roboto Technologies, Inc.

import datetime
from distutils.version import StrictVersion
import enum
import importlib.metadata
import os
import sys
from typing import Optional

import pydantic

from roboto.http import HttpClient
from roboto.profile import RobotoProfile
from roboto.time import utcnow


class CLIState(pydantic.BaseModel):
    last_checked_version: Optional[datetime.datetime]
    last_version: str = "0.0.0"
    out_of_date: bool = True


class AnsiColor(str, enum.Enum):
    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    BLUE = "\033[0;34m"
    END = "\033[0m"


def check_last_update(profile: RobotoProfile):
    roboto_tmp_dir = profile.config_dir / "tmp"
    roboto_tmp_dir.mkdir(parents=True, exist_ok=True)
    cli_state_file = roboto_tmp_dir / "cli_state.json"

    last_version = None
    version = importlib.metadata.version("roboto")

    state = CLIState(last_checked_version=None)
    if cli_state_file.is_file():
        state = CLIState.parse_file(cli_state_file)
        last_version = state.last_version

    if (
        state.last_checked_version is None
        or version != last_version
        or state.out_of_date is None
        or state.out_of_date is True
        or (utcnow() - datetime.timedelta(hours=1)) > state.last_checked_version
    ):
        http = HttpClient()

        versions = list(
            http.get(url="https://pypi.org/pypi/roboto-sdk/json")
            .from_json(json_path=["releases"])
            .keys()
        )
        versions.sort(key=StrictVersion)
        latest = versions[-1]

        state.last_checked_version = utcnow()
        state.last_version = version
        state.out_of_date = version != latest

        cli_state_file.write_text(state.json())

        suppress_message = (
            os.getenv("ROBOTO_CLI_SUPPRESS_UPGRADE_PROMPT", "false").lower() != "false"
        )

        if state.out_of_date and not suppress_message:
            notice = f"{AnsiColor.BLUE}[notice]{AnsiColor.END}"
            print(
                f"\n{notice} A new release of roboto_sdk is available: "
                + f"{AnsiColor.RED + version + AnsiColor.END} -> {AnsiColor.GREEN + latest + AnsiColor.END}\n"
                + f"{notice} To update, run: {AnsiColor.GREEN}pip install --upgrade roboto-sdk{AnsiColor.END}",
                file=sys.stderr,
            )



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ..domain.actions import (
    ActionDelegate,
    InvocationDelegate,
)
from ..domain.datasets import DatasetDelegate
from ..domain.files import FileDelegate
from ..domain.orgs import OrgDelegate
from ..domain.tokens import TokenDelegate
from ..domain.triggers import TriggerDelegate
from ..domain.users import UserDelegate
from ..http import HttpClient


class CLIContext:
    __roboto_service_base_url: Optional[str]
    __http: Optional[HttpClient]
    actions: ActionDelegate
    datasets: DatasetDelegate
    files: FileDelegate
    invocations: InvocationDelegate
    orgs: OrgDelegate
    tokens: TokenDelegate
    triggers: TriggerDelegate
    users: UserDelegate
    extensions: dict[str, Any]

    @property
    def roboto_service_base_url(self) -> str:
        if self.__roboto_service_base_url is None:
            raise ValueError("roboto_service_base_url is unset")

        return self.__roboto_service_base_url

    @roboto_service_base_url.setter
    def roboto_service_base_url(self, roboto_service_base_url: str) -> None:
        self.__roboto_service_base_url = roboto_service_base_url

    @property
    def http(self) -> HttpClient:
        # Necessary since http is lazy set after parsing args
        if self.__http is None:
            raise ValueError("Unset HTTP client!")

        return self.__http

    @http.setter
    def http(self, http: HttpClient) -> None:
        self.__http = http



#  Copyright (c) 2023 Roboto Technologies, Inc.

RESOURCE_OWNER_OVERRIDE_HEADER = "X-Roboto-Resource-Owner-Id"
"""Header to specify the organization that owns the resource being accessed."""

RESOURCE_OWNER_OVERRIDE_QUERY_PARAM = "robotoResourceOwnerId"
"""Query parameter to specify the organization that owns the resource being accessed."""

ORG_OVERRIDE_HEADER = "X-Roboto-Org-Id"
"""Header to specify the organization that the user is acting on behalf of."""

ORG_OVERRIDE_QUERY_PARAM = "robotoOrgId"
"""Query parameter to specify the organization that the user is acting on behalf of."""

USER_OVERRIDE_HEADER = "X-Roboto-User-Id"
"""Header to specify the user that is performing the REST operation."""

USER_OVERRIDE_QUERY_PARAM = "robotoUserId"
""""Query parameter to specify the user that is performing the REST operation."""



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import json
import time

from ...domain.actions import (
    Invocation,
    InvocationStatus,
)
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def status(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    if not args.tail:
        invocation = Invocation.from_id(
            args.invocation_id,
            invocation_delegate=context.invocations,
            org_id=args.org,
        )
        print(
            json.dumps(
                [
                    status_record.to_presentable_dict()
                    for status_record in invocation.status_log
                ],
                indent=4,
            )
        )
        return

    printed: set[InvocationStatus] = set()
    terminal_statuses = {
        InvocationStatus.Completed,
        InvocationStatus.Failed,
        InvocationStatus.Deadly,
    }
    while len(printed & terminal_statuses) == 0:
        invocation = Invocation.from_id(
            args.invocation_id,
            invocation_delegate=context.invocations,
            org_id=args.org,
        )

        status_records_to_print = [
            status_record
            for status_record in invocation.status_log
            if status_record.status not in printed
        ]
        if status_records_to_print:
            for status_record in status_records_to_print:
                printed.add(status_record.status)
                print(json.dumps(status_record.to_presentable_dict(), indent=4))

        time.sleep(1)


def status_parser(parser: argparse.ArgumentParser):
    parser.add_argument("invocation_id")
    parser.add_argument("--tail", required=False, action="store_true")
    add_org_arg(parser=parser)


status_command = RobotoCommand(
    name="status",
    logic=status,
    setup_parser=status_parser,
    command_kwargs={"help": "Get invocation status."},
)



import http
import json
import logging
import typing
import urllib.error
import urllib.parse
import urllib.request
import urllib.response

import tenacity
import tenacity.wait

from ..logging import LOGGER_NAME
from ..serde import safe_dict_drill

logger = logging.getLogger(LOGGER_NAME)

DEFAULT_HEADERS = object()


def is_expected_to_be_transient(exc: typing.Any) -> bool:
    if isinstance(exc, urllib.error.HTTPError):
        try:
            return http.HTTPStatus(int(exc.code)) in (
                http.HTTPStatus.TOO_MANY_REQUESTS,
                http.HTTPStatus.SERVICE_UNAVAILABLE,
            )
        except ValueError:
            return False

    # Must come after the isinstance check of HTTPError,
    # which is a subclass of URLError
    if isinstance(exc, urllib.error.URLError):
        # DNS error, probably transient
        return "Temporary failure in name resolution" in str(exc.reason)

    return False


def default_retry_wait(_exc: typing.Optional[BaseException]) -> float:
    return 1.0


class HttpError(Exception):
    __http_exc: urllib.error.HTTPError
    __msg: typing.Any = None
    __status: typing.Optional[http.HTTPStatus] = None

    def __init__(self, exc: urllib.error.HTTPError) -> None:
        super().__init__()
        self.__http_exc = exc

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.msg!r}, {self.status!r})"

    def __str__(self) -> str:
        return f"{self.msg!r}"

    @property
    def status(self) -> typing.Optional[http.HTTPStatus]:
        if self.__status is None:
            try:
                self.__status = http.HTTPStatus(int(self.__http_exc.code))
            except ValueError:
                self.__status = None
        return self.__status

    @property
    def msg(self) -> typing.Any:
        if self.__msg is None:
            decoded = self.__http_exc.read().decode("utf-8")
            try:
                self.__msg = json.loads(decoded)
            except json.JSONDecodeError:
                self.__msg = decoded
        return self.__msg

    @property
    def headers(self) -> dict:
        return dict(self.__http_exc.headers.items())


class ClientError(HttpError):
    pass


class ServerError(HttpError):
    pass


class HttpResponse:
    __response: urllib.response.addinfourl

    def __init__(self, response: urllib.response.addinfourl) -> None:
        super().__init__()
        self.__response = response

    @property
    def readable_response(self) -> urllib.response.addinfourl:
        return self.__response

    @property
    def status(self) -> http.HTTPStatus:
        status_code = self.__response.status
        if status_code is None:
            raise RuntimeError("Response has no status code")
        return http.HTTPStatus(int(status_code))

    @property
    def headers(self) -> typing.Optional[dict[str, str]]:
        return dict(self.__response.headers.items())

    def from_json(self, json_path: typing.Optional[list[str]] = None) -> typing.Any:
        with self.__response:
            unmarsalled = json.loads(self.__response.read().decode("utf-8"))
            if json_path is None:
                return unmarsalled

            return safe_dict_drill(unmarsalled, json_path)

    def from_string(self):
        with self.__response:
            return self.__response.read().decode("utf-8")


RetryWaitFn = typing.Callable[[typing.Optional[BaseException]], float]


class HttpRequest:
    url: str
    method: str
    headers: dict
    retry_wait: RetryWaitFn
    data: typing.Any = None

    def __init__(
        self,
        url: str,
        method: str = "GET",
        headers: typing.Optional[dict[str, str]] = None,
        data: typing.Any = None,
        retry_wait: typing.Optional[RetryWaitFn] = None,
    ):
        self.url = url
        self.method = method
        self.headers = headers if headers is not None else {}
        self.data = data
        self.retry_wait = retry_wait if retry_wait is not None else default_retry_wait

    def __repr__(self) -> str:
        return f"HttpRequest(url={self.url}, method={self.method}, headers={self.headers}, data={self.data})"

    @property
    def body(self) -> typing.Optional[bytes]:
        if self.data is None:
            return None

        if isinstance(self.data, bytes):
            return self.data

        if isinstance(self.data, str):
            return self.data.encode("utf-8")

        return json.dumps(self.data).encode("utf-8")

    @property
    def hostname(self) -> str:
        parsed_url = urllib.parse.urlparse(self.url)
        return (
            parsed_url.hostname
            if parsed_url.hostname is not None
            else parsed_url.netloc
        )

    def append_headers(self, headers: dict[str, str]) -> None:
        self.headers.update(headers)


HttpRequestDecorator = typing.Callable[[HttpRequest], HttpRequest]


class HttpClient:
    __base_headers: dict[str, str]
    __default_auth: typing.Optional[HttpRequestDecorator]
    __default_endpoint: typing.Optional[str]

    def __init__(
        self,
        base_headers: typing.Optional[dict[str, str]] = None,
        default_endpoint: typing.Optional[str] = None,
        default_auth: typing.Optional[HttpRequestDecorator] = None,
    ):
        self.__base_headers = base_headers if base_headers is not None else {}
        self.__default_auth = default_auth
        self.__default_endpoint = default_endpoint

    def delete(
        self, url, data: typing.Any = None, headers: typing.Optional[dict] = None
    ) -> HttpResponse:
        request = HttpRequest(url=url, method="DELETE", data=data, headers=headers)
        return self.__request(request)

    def get(
        self,
        url: str,
        headers: typing.Optional[dict] = None,
        retry_wait: RetryWaitFn = default_retry_wait,
    ) -> HttpResponse:
        request = HttpRequest(
            url=url, method="GET", headers=headers, retry_wait=retry_wait
        )
        return self.__request(request)

    def post(
        self, url, data: typing.Any = None, headers: typing.Optional[dict] = None
    ) -> HttpResponse:
        request = HttpRequest(url=url, method="POST", data=data, headers=headers)
        return self.__request(request)

    def patch(
        self, url, data: typing.Any = None, headers: typing.Optional[dict] = None
    ) -> HttpResponse:
        request = HttpRequest(url=url, method="PATCH", data=data, headers=headers)
        return self.__request(request)

    def put(
        self, url, data: typing.Any = None, headers: typing.Optional[dict] = None
    ) -> HttpResponse:
        request = HttpRequest(url=url, method="PUT", data=data, headers=headers)
        return self.__request(request)

    def url(self, path: str) -> str:
        if self.__default_endpoint is None:
            raise ValueError(
                "HttpClient.url called for client with no default endpoint. Just pass the URL clear!"
            )
        return f"{self.__default_endpoint}/{path}"

    def __request(self, request_ctx: HttpRequest) -> HttpResponse:
        if self.__default_auth is not None:
            request_ctx = self.__default_auth(request_ctx)

        logger.debug("%s", request_ctx)

        headers = self.__base_headers.copy()
        if request_ctx.headers:
            headers.update(request_ctx.headers)

        try:
            for attempt in tenacity.Retrying(
                retry=tenacity.retry_if_exception(is_expected_to_be_transient),
                stop=tenacity.stop_after_attempt(3),
                reraise=True,
                wait=self._wait(request_ctx.retry_wait),
            ):
                with attempt:
                    request = urllib.request.Request(
                        request_ctx.url, method=request_ctx.method
                    )

                    req_body = request_ctx.body
                    if req_body is not None:
                        request.data = req_body

                    for key, value in headers.items():
                        request.add_header(key, value)

                    return HttpResponse(urllib.request.urlopen(request))
        except urllib.error.HTTPError as exc:
            status_code = exc.code
            if status_code > 399 and status_code < 500:
                raise ClientError(exc) from None
            elif status_code > 499:
                raise ServerError(exc) from None
            else:
                raise HttpError(exc) from None
        except Exception:
            raise

        raise RuntimeError("Unreachable")

    def _wait(self, waiter: RetryWaitFn) -> tenacity.wait.wait_base:
        class Waiter(tenacity.wait.wait_base):
            def __call__(self, retry_state: tenacity.RetryCallState) -> float:
                outcome = retry_state.outcome
                if outcome is None:
                    return waiter(None)
                return waiter(outcome.exception())

        return Waiter()



import collections.abc
import datetime
from typing import Any, Optional, Union

from ...query import QuerySpecification
from ...sentinels import NotSet, NotSetType
from ...serde import pydantic_jsonable_dict
from ...updates import MetadataChangeset
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .action_delegate import ActionDelegate
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionRecord,
    ActionReference,
)
from .invocation import Invocation
from .invocation_delegate import (
    InvocationDelegate,
)
from .invocation_record import (
    InvocationDataSourceType,
    InvocationSource,
)


class Action:
    __action_delegate: ActionDelegate
    __invocation_delegate: InvocationDelegate
    __record: ActionRecord

    @classmethod
    def create(
        cls,
        name: str,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        parameters: Optional[list[ActionParameter]] = None,
        uri: Optional[str] = None,
        inherits: Optional[ActionReference] = None,
        compute_requirements: Optional[ComputeRequirements] = None,
        container_parameters: Optional[ContainerParameters] = None,
        description: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        tags: Optional[list[str]] = None,
        created_by: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> "Action":
        record = action_delegate.create_action(
            name,
            parameters,
            uri,
            inherits,
            compute_requirements,
            container_parameters,
            description,
            metadata,
            tags,
            created_by,
            org_id,
        )
        return cls(record, action_delegate, invocation_delegate)

    @classmethod
    def from_name(
        cls,
        name: str,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        org_id: Optional[str] = None,
        digest: Optional[str] = None,
        action_owner_id: Optional[str] = None,
    ) -> "Action":
        record = action_delegate.get_action_by_primary_key(
            name, org_id, digest=digest, action_owner_id=action_owner_id
        )
        return cls(record, action_delegate, invocation_delegate)

    @classmethod
    def query(
        cls,
        query: QuerySpecification,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        accessibility: Accessibility = Accessibility.Organization,
        org_id: Optional[str] = None,
    ) -> collections.abc.Generator["Action", None, None]:
        known = set(ActionRecord.__fields__.keys())
        actual = set()
        for field in query.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of Action"
                if plural
                else "is not a known attribute of Action"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        query_method = (
            action_delegate.query_actions_on_action_hub
            if accessibility == Accessibility.ActionHub
            else action_delegate.query_actions
        )
        paginated_results = query_method(query, org_id=org_id)
        while True:
            for record in paginated_results.items:
                yield cls(record, action_delegate, invocation_delegate)
            if paginated_results.next_token:
                query.after = paginated_results.next_token
                paginated_results = query_method(query, org_id=org_id)
            else:
                break

    def __init__(
        self,
        record: ActionRecord,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
    ) -> None:
        self.__action_delegate = action_delegate
        self.__invocation_delegate = invocation_delegate
        self.__record = record

    @property
    def accessibility(self) -> Accessibility:
        return self.__record.accessibility

    @property
    def compute_requirements(self) -> Optional[ComputeRequirements]:
        return self.__record.compute_requirements

    @property
    def container_parameters(self) -> Optional[ContainerParameters]:
        return self.__record.container_parameters

    @property
    def digest(self) -> str:
        return (
            self.__record.digest
            if self.__record.digest
            else self.__record.compute_digest()
        )

    @property
    def inherits_from(self) -> Optional[ActionReference]:
        return self.__record.inherits

    @property
    def last_modified(self) -> datetime.datetime:
        return self.__record.modified

    @property
    def name(self) -> str:
        return self.__record.name

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def parameters(self) -> list[ActionParameter]:
        return [param.copy() for param in self.__record.parameters]

    @property
    def record(self) -> ActionRecord:
        return self.__record

    @property
    def uri(self) -> Optional[str]:
        return self.__record.uri

    def set_accessibility(self, accessibility: Accessibility) -> None:
        self.__record = self.__action_delegate.set_accessibility(
            self.__record, accessibility
        )

    def delete(self) -> None:
        self.__action_delegate.delete_action(self.__record)

    def invoke(
        self,
        input_data: list[str],
        data_source_id: str,
        data_source_type: InvocationDataSourceType,
        invocation_source: InvocationSource,
        invocation_source_id: Optional[str] = None,
        parameter_values: Optional[dict[str, Any]] = None,
        compute_requirement_overrides: Optional[ComputeRequirements] = None,
        container_parameter_overrides: Optional[ContainerParameters] = None,
        idempotency_id: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> Invocation:
        record = self.__invocation_delegate.create_invocation(
            action_record=self.__record,
            parameter_values=parameter_values if parameter_values else dict(),
            input_data=input_data,
            data_source_id=data_source_id,
            data_source_type=data_source_type,
            invocation_source=invocation_source,
            invocation_source_id=invocation_source_id,
            compute_requirement_overrides=compute_requirement_overrides,
            container_parameter_overrides=container_parameter_overrides,
            idempotency_id=idempotency_id,
            org_id=org_id,
        )
        return Invocation(
            record,
            self.__invocation_delegate,
        )

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)

    def update(
        self,
        compute_requirements: Union[ComputeRequirements, NotSetType] = NotSet,
        container_parameters: Union[ContainerParameters, NotSetType] = NotSet,
        description: Optional[Union[str, NotSetType]] = NotSet,
        metadata_changeset: Union[MetadataChangeset, NotSetType] = NotSet,
        parameter_changeset: Union[ActionParameterChangeset, NotSetType] = NotSet,
        uri: Union[str, NotSetType] = NotSet,
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> None:
        updated = self.__action_delegate.update(
            self.__record,
            compute_requirements=compute_requirements,
            container_parameters=container_parameters,
            description=description,
            metadata_changeset=metadata_changeset,
            parameter_changeset=parameter_changeset,
            uri=uri,
            updated_by=updated_by,
        )
        self.__record = updated



import typing

import pydantic

from ...updates import (
    MetadataChangeset,
    UpdateCondition,
)
from .record import Administrator, StorageLocation


class CreateDatasetRequest(pydantic.BaseModel):
    administrator: Administrator
    storage_location: StorageLocation
    description: typing.Optional[str]
    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    tags: list[str] = pydantic.Field(default_factory=list)


class QueryDatasetsRequest(pydantic.BaseModel):
    filters: dict[str, typing.Any] = pydantic.Field(default_factory=dict)

    class Config:
        extra = "forbid"


class UpdateDatasetRequest(pydantic.BaseModel):
    metadata_changeset: typing.Optional[MetadataChangeset] = None
    description: typing.Optional[str] = None
    conditions: typing.Optional[list[UpdateCondition]] = None



from .domain import (
    RobotoConditionException,
    RobotoConflictException,
    RobotoDomainException,
    RobotoHttpExceptionParse,
    RobotoIllegalArgumentException,
    RobotoInvalidRequestException,
    RobotoLimitExceededException,
    RobotoNoOrgProvidedException,
    RobotoNotFoundException,
    RobotoServiceException,
    RobotoUnauthorizedException,
    RobotoUnknownOperationException,
)

__all__ = [
    "RobotoConditionException",
    "RobotoConflictException",
    "RobotoDomainException",
    "RobotoHttpExceptionParse",
    "RobotoIllegalArgumentException",
    "RobotoInvalidRequestException",
    "RobotoLimitExceededException",
    "RobotoNotFoundException",
    "RobotoNoOrgProvidedException",
    "RobotoServiceException",
    "RobotoUnauthorizedException",
    "RobotoUnknownOperationException",
]



#  Copyright (c) 2023 Roboto Technologies, Inc.
from .action import Action
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
    ExecutorContainer,
)
from .action_delegate import ActionDelegate
from .action_http_delegate import (
    ActionHttpDelegate,
)
from .action_http_resources import (
    CreateActionRequest,
    QueryActionsRequest,
    SetActionAccessibilityRequest,
    UpdateActionRequest,
)
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionRecord,
    ActionReference,
)
from .invocation import Invocation
from .invocation_delegate import (
    InvocationDelegate,
)
from .invocation_http_delegate import (
    InvocationHttpDelegate,
)
from .invocation_http_resources import (
    CreateInvocationRequest,
    QueryInvocationsRequest,
    SetLogsLocationRequest,
    UpdateInvocationStatus,
)
from .invocation_record import (
    ActionProvenance,
    ExecutableProvenance,
    InvocationDataSource,
    InvocationDataSourceType,
    InvocationProvenance,
    InvocationRecord,
    InvocationSource,
    InvocationStatus,
    InvocationStatusRecord,
    LogRecord,
    LogsLocation,
    SourceProvenance,
)

__all__ = (
    "Accessibility",
    "Action",
    "ActionDelegate",
    "ActionHttpDelegate",
    "ActionParameter",
    "ActionParameterChangeset",
    "ActionProvenance",
    "ActionRecord",
    "ActionReference",
    "ComputeRequirements",
    "ContainerParameters",
    "CreateActionRequest",
    "CreateInvocationRequest",
    "ExecutableProvenance",
    "ExecutorContainer",
    "Invocation",
    "InvocationDataSource",
    "InvocationDataSourceType",
    "InvocationDelegate",
    "InvocationHttpDelegate",
    "InvocationProvenance",
    "InvocationRecord",
    "InvocationSource",
    "InvocationStatus",
    "InvocationStatusRecord",
    "LogsLocation",
    "LogRecord",
    "QueryActionsRequest",
    "QueryInvocationsRequest",
    "SetActionAccessibilityRequest",
    "SetLogsLocationRequest",
    "SourceProvenance",
    "UpdateActionRequest",
    "UpdateInvocationStatus",
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

from ..command import RobotoCommandSet
from .create import create_command
from .delete_dataset import delete_dataset_command
from .delete_files import delete_files_command
from .download_files import download_files_command
from .list_files import list_files_command
from .search import search_command
from .show import show_command
from .update import update_command
from .upload_files import upload_files_command

commands = [
    create_command,
    delete_dataset_command,
    delete_files_command,
    download_files_command,
    list_files_command,
    show_command,
    search_command,
    update_command,
    upload_files_command,
]

command_set = RobotoCommandSet(
    name="datasets",
    help="Get, create, update, delete, and query datasets. Upload to and download from them.",
    commands=commands,
)



#  Copyright (c) 2023 Roboto Technologies, Inc.


from typing import Optional

import pydantic


class CreateTokenRequest(pydantic.BaseModel):
    expiry_days: int
    name: str
    description: Optional[str] = None


class ValidateTokenResponse(pydantic.BaseModel):
    valid: bool



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import json

from ...domain.actions import Action
from ..command import RobotoCommand
from ..common_args import (
    add_action_reference_arg,
    add_org_arg,
)
from ..context import CLIContext


def show(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    action = Action.from_name(
        name=args.action.name,
        action_delegate=context.actions,
        invocation_delegate=context.invocations,
        org_id=args.org,
        digest=args.action.digest,
        action_owner_id=args.action.owner,
    )
    print(json.dumps(action.to_dict(), indent=4))


def show_parser(parser: argparse.ArgumentParser):
    add_action_reference_arg(parser)
    add_org_arg(parser=parser)


show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_parser,
    command_kwargs={"help": "Get an Action."},
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

from .http_delegate import UserHttpDelegate
from .user import User
from .user_delegate import UserDelegate
from .user_http_resources import UpdateUserRequest
from .user_record import UserRecord

__all__ = [
    "User",
    "UserDelegate",
    "UserRecord",
    "UserHttpDelegate",
    "UpdateUserRequest",
]



#  Copyright (c) 2023 Roboto Technologies, Inc.

from argparse import HelpFormatter
from operator import attrgetter


class SortingHelpFormatter(HelpFormatter):
    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter("option_strings"))
        super(SortingHelpFormatter, self).add_arguments(actions)



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import sys

from ...domain.datasets import Dataset
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def list_files(args, context: CLIContext, parser: argparse.ArgumentParser):
    record = Dataset.from_id(
        args.dataset_id, context.datasets, context.files, org_id=args.org
    )

    for f in record.list_files():
        sys.stdout.write(f"{f.relative_path}\n")


def list_files_setup_parser(parser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )
    add_org_arg(parser)


list_files_command = RobotoCommand(
    name="list-files",
    logic=list_files,
    setup_parser=list_files_setup_parser,
    command_kwargs={"help": "Lists files for a specific dataset."},
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

from .profile import (
    MalformedProfileException,
    MissingProfileException,
    RobotoProfile,
    RobotoProfileEntry,
)

__all__ = [
    "RobotoProfile",
    "RobotoProfileEntry",
    "MissingProfileException",
    "MalformedProfileException",
]



#  Copyright (c) 2023 Roboto Technologies, Inc.

from .delegate import TokenDelegate
from .http_delegate import TokenHttpDelegate
from .http_resources import (
    CreateTokenRequest,
    ValidateTokenResponse,
)
from .record import TokenContext, TokenRecord
from .token import Token

__all__ = [
    "CreateTokenRequest",
    "Token",
    "TokenDelegate",
    "TokenHttpDelegate",
    "TokenRecord",
    "TokenContext",
    "ValidateTokenResponse",
]



import collections.abc
import decimal
import enum
import json
import typing

import pydantic

from ..serde import safe_dict_drill


class Comparator(str, enum.Enum):
    """The comparator to use when comparing a field to a value."""

    Equals = "EQUALS"
    NotEquals = "NOT_EQUALS"
    GreaterThan = "GREATER_THAN"
    GreaterThanOrEqual = "GREATER_THAN_OR_EQUAL"
    LessThan = "LESS_THAN"
    LessThanOrEqual = "LESS_THAN_OR_EQUAL"
    Contains = "CONTAINS"
    NotContains = "NOT_CONTAINS"
    IsNull = "IS_NULL"
    IsNotNull = "IS_NOT_NULL"
    Exists = "EXISTS"
    NotExists = "NOT_EXISTS"
    BeginsWith = "BEGINS_WITH"


class Condition(pydantic.BaseModel):
    """A filter for any arbitrary attribute for a Roboto resource."""

    field: str
    comparator: Comparator
    value: typing.Optional[typing.Union[str, bool, int, float, decimal.Decimal]] = None

    @pydantic.validator("value")
    def parse(cls, v):
        if v is None:
            return None
        try:
            return json.loads(v)
        except json.decoder.JSONDecodeError:
            return v

    def matches(self, target: dict) -> bool:
        value = safe_dict_drill(target, self.field.split("."))

        if self.comparator in [Comparator.NotExists, Comparator.IsNull]:
            return value is None

        if self.comparator in [Comparator.Exists, Comparator.IsNotNull]:
            return value is not None

        # We need the value for everything else
        if value is None:
            return False

        if isinstance(value, str) and not isinstance(self.value, str):
            if isinstance(self.value, int):
                value = int(value)
            elif isinstance(self.value, float):
                value = float(value)
            elif isinstance(self.value, bool):
                value = value.lower() == "true"
            elif isinstance(self.value, decimal.Decimal):
                value = decimal.Decimal.from_float(float(value))

        if self.comparator is Comparator.Equals:
            return value == self.value

        if self.comparator is Comparator.NotEquals:
            return value != self.value

        if self.comparator is Comparator.GreaterThan:
            return value > self.value

        if self.comparator is Comparator.GreaterThanOrEqual:
            return value >= self.value

        if self.comparator is Comparator.LessThan:
            return value < self.value

        if self.comparator is Comparator.LessThanOrEqual:
            return value <= self.value

        if self.comparator is Comparator.Contains:
            return isinstance(value, list) and self.value in value

        if self.comparator is Comparator.NotContains:
            return isinstance(value, list) and self.value not in value

        if self.comparator is Comparator.BeginsWith:
            if not isinstance(value, str) or not isinstance(self.value, str):
                return False

            return value.startswith(self.value)

        return False

    def __str__(self):
        return self.json()

    def __repr__(self):
        return self.json()


class ConditionOperator(str, enum.Enum):
    """The operator to use when combining multiple conditions."""

    And = "AND"
    Or = "OR"
    Not = "NOT"


class ConditionGroup(pydantic.BaseModel):
    """A group of conditions that are combined together."""

    operator: ConditionOperator
    conditions: collections.abc.Sequence[typing.Union[Condition, "ConditionGroup"]]

    def matches(self, target: dict):
        inner_matches = map(lambda x: x.matches(target), self.conditions)

        if self.operator is ConditionOperator.And:
            return all(inner_matches)

        if self.operator is ConditionOperator.Or:
            return any(inner_matches)

        if self.operator is ConditionOperator.Not:
            return not any(inner_matches)

        return False

    @pydantic.validator("conditions")
    def validate_conditions(
        cls, v: collections.abc.Sequence[typing.Union[Condition, "ConditionGroup"]]
    ):
        if len(v) == 0:
            raise ValueError(
                "At least one condition must be provided to a ConditionGroup, got 0!"
            )

        return v

    def __str__(self):
        return self.json()

    def __repr__(self):
        return self.json()


ConditionType = typing.Union[Condition, ConditionGroup]



#  Copyright (c) 2023 Roboto Technologies, Inc.
from typing import Any, Optional, Union

from roboto.sentinels import (
    NotSet,
    NotSetType,
    is_set,
)

from ...exceptions import RobotoHttpExceptionParse
from ...http import (
    HttpClient,
    PaginatedList,
    roboto_headers,
)
from ...query import (
    ConditionType,
    QuerySpecification,
)
from ...serde import pydantic_jsonable_dict
from ..actions import (
    ComputeRequirements,
    ContainerParameters,
)
from .http_resources import (
    CreateTriggerRequest,
    UpdateTriggerRequest,
)
from .trigger_delegate import TriggerDelegate
from .trigger_record import (
    TriggerForEachPrimitive,
    TriggerRecord,
)


class TriggerHttpDelegate(TriggerDelegate):
    __http_client: HttpClient

    def __init__(self, http_client: HttpClient):
        super().__init__()
        self.__http_client = http_client

    def create_trigger(
        self,
        name: str,
        action_name: str,
        required_inputs: list[str],
        for_each: TriggerForEachPrimitive,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,  # A Roboto user_id
        compute_requirement_overrides: Optional[ComputeRequirements] = None,
        container_parameter_overrides: Optional[ContainerParameters] = None,
        condition: Optional[ConditionType] = None,
        additional_inputs: Optional[list[str]] = None,
        parameter_values: Optional[dict[str, Any]] = None,
        action_owner_id: Optional[str] = None,
        action_digest: Optional[str] = None,
    ) -> TriggerRecord:
        url = self.__http_client.url("v1/triggers")
        headers = roboto_headers(
            org_id=org_id,
            user_id=created_by,
            additional_headers={"Content-Type": "application/json"},
        )

        request_body = CreateTriggerRequest(
            name=name,
            action_name=action_name,
            action_owner_id=action_owner_id,
            action_digest=action_digest,
            required_inputs=required_inputs,
            compute_requirement_overrides=compute_requirement_overrides,
            container_parameter_overrides=container_parameter_overrides,
            condition=condition,
            for_each=for_each,
            additional_inputs=additional_inputs,
            parameter_values=parameter_values,
        )

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url, data=pydantic_jsonable_dict(request_body), headers=headers
            )

        return TriggerRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_trigger_by_primary_key(
        self, name: str, org_id: Optional[str] = None
    ) -> TriggerRecord:
        url = self.__http_client.url(f"v1/triggers/{name}")
        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url, headers=headers)

        return TriggerRecord.parse_obj(response.from_json(json_path=["data"]))

    def query_triggers(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[TriggerRecord]:
        url = self.__http_client.url("v1/triggers/query")
        post_body = pydantic_jsonable_dict(query, exclude_none=True)
        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url,
                data=post_body,
                headers=roboto_headers(
                    org_id=org_id,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )

        unmarshalled = response.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                TriggerRecord.parse_obj(trigger) for trigger in unmarshalled["items"]
            ],
            next_token=unmarshalled.get("next_token"),
        )

    def delete_trigger(self, name: str, org_id: str) -> None:
        url = self.__http_client.url(f"v1/triggers/{name}")
        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            self.__http_client.delete(url=url, headers=headers)

    def update_trigger(
        self,
        name: str,
        org_id: str,
        updated_by: Optional[str] = None,
        action_name: Union[str, NotSetType] = NotSet,
        action_owner_id: Union[str, NotSetType] = NotSet,
        action_digest: Optional[Union[str, NotSetType]] = NotSet,
        required_inputs: Union[list[str], NotSetType] = NotSet,
        for_each: Union[TriggerForEachPrimitive, NotSetType] = NotSet,
        enabled: Union[bool, NotSetType] = NotSet,
        additional_inputs: Optional[Union[list[str], NotSetType]] = NotSet,
        compute_requirement_overrides: Optional[
            Union[ComputeRequirements, NotSetType]
        ] = NotSet,
        container_parameter_overrides: Optional[
            Union[ContainerParameters, NotSetType]
        ] = NotSet,
        condition: Optional[Union[ConditionType, NotSetType]] = NotSet,
        parameter_values: Optional[Union[dict[str, Any], NotSetType]] = NotSet,
    ) -> TriggerRecord:
        url = self.__http_client.url(f"v1/triggers/{name}")
        headers = roboto_headers(org_id=org_id)
        updates = {
            "action_name": action_name,
            "action_owner_id": action_owner_id,
            "action_digest": action_digest,
            "required_inputs": required_inputs,
            "for_each": for_each,
            "enabled": enabled,
            "additional_inputs": additional_inputs,
            "compute_requirement_overrides": compute_requirement_overrides,
            "container_parameter_overrides": container_parameter_overrides,
            "condition": condition,
            "parameter_values": parameter_values,
        }
        request_body = UpdateTriggerRequest.parse_obj(
            {k: v for k, v in updates.items() if is_set(v)}
        )
        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url=url,
                headers=headers,
                data=pydantic_jsonable_dict(request_body, exclude_unset=True),
            )

        return TriggerRecord.parse_obj(response.from_json(json_path=["data"]))



from .constants import (
    ORG_OVERRIDE_HEADER,
    ORG_OVERRIDE_QUERY_PARAM,
    RESOURCE_OWNER_OVERRIDE_HEADER,
    RESOURCE_OWNER_OVERRIDE_QUERY_PARAM,
    USER_OVERRIDE_HEADER,
    USER_OVERRIDE_QUERY_PARAM,
)
from .headers import roboto_headers
from .http_client import (
    ClientError,
    HttpClient,
    HttpError,
    ServerError,
)
from .request_decorators import (
    LocalAuthDecorator,
    PATAuthDecorator,
    SigV4AuthDecorator,
)
from .response import (
    PaginatedList,
    PaginationToken,
    PaginationTokenEncoding,
    PaginationTokenScheme,
    StreamedList,
)
from .testing_util import FakeHttpResponseFactory

__all__ = (
    "roboto_headers",
    "ClientError",
    "FakeHttpResponseFactory",
    "HttpClient",
    "HttpError",
    "LocalAuthDecorator",
    "PaginatedList",
    "PaginationToken",
    "PaginationTokenEncoding",
    "PaginationTokenScheme",
    "PATAuthDecorator",
    "ServerError",
    "SigV4AuthDecorator",
    "StreamedList",
    "ORG_OVERRIDE_HEADER",
    "ORG_OVERRIDE_QUERY_PARAM",
    "RESOURCE_OWNER_OVERRIDE_HEADER",
    "RESOURCE_OWNER_OVERRIDE_QUERY_PARAM",
    "USER_OVERRIDE_HEADER",
    "USER_OVERRIDE_QUERY_PARAM",
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

import collections.abc
from typing import Any, Optional, Union

from roboto.sentinels import (
    NotSet,
    NotSetType,
)

from ...domain.actions import (
    Action,
    ActionDelegate,
    ComputeRequirements,
    ContainerParameters,
    Invocation,
    InvocationDataSource,
    InvocationDelegate,
    InvocationSource,
)
from ...query import (
    ConditionType,
    QuerySpecification,
)
from ...serde import pydantic_jsonable_dict
from .trigger_delegate import TriggerDelegate
from .trigger_record import (
    TriggerForEachPrimitive,
    TriggerRecord,
)


class Trigger:
    __record: TriggerRecord
    __action_delegate: ActionDelegate
    __invocation_delegate: InvocationDelegate
    __trigger_delegate: TriggerDelegate

    @classmethod
    def create(
        cls,
        name: str,
        action_name: str,
        required_inputs: list[str],
        for_each: TriggerForEachPrimitive,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        trigger_delegate: TriggerDelegate,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,
        compute_requirement_overrides: Optional[ComputeRequirements] = None,
        container_parameter_overrides: Optional[ContainerParameters] = None,
        condition: Optional[ConditionType] = None,
        additional_inputs: Optional[list[str]] = None,
        parameter_values: Optional[dict[str, Any]] = None,
        action_owner_id: Optional[str] = None,
        action_digest: Optional[str] = None,
    ) -> "Trigger":
        """
        Creates an executor trigger, which automatically invokes an action on every new dataset that meets
        some acceptance criteria.

        Args:
            name: A human-readable name for this trigger. Trigger names must be unique within each
                organization, though collisions are fine across different organizations.
            action_name: The name of an executor action to run.
                If an action with the specified name is not found, the trigger will not be created.
                If the action bound to this trigger is ever deleted, the trigger will be deleted along with it.
            required_inputs: A list of gitignore path patterns that describe a set of files required to
                invoke an action described by this trigger against a given dataset.
                An action will be invoked if at least one file is uploaded that matches each listed condition.
                Once invoked, this list of path patterns will also be used to determine which files from the dataset
                to make available to the action at runtime (i.e., downloaded into $INPUT_DIR).
                If you want to make the entire dataset available to the action, add a condition for "**/*" to the
                end of required_inputs.
            org_id: The ID of the organization the user is making this request on behalf of. If the user is
                only a member of one organization, this parameter will be set implicitly.
            compute_requirement_overrides: Optional overrides of the compute parameters specified by the
                action.
            container_parameter_overrides: Optional overrides of the container parameters specified by the
                action.
            action_delegate: An abstraction object for performing actions against the actions API.
            invocation_delegate: An abstraction object for performing actions against the invocations API.
            trigger_delegate: An abstraction object for performing actions against the triggers API.

        Returns:
            Trigger: A reference to a Trigger entity object which allows the user to perform additional operations
            on the newly created Trigger.
        """
        record = trigger_delegate.create_trigger(
            name=name,
            org_id=org_id,
            action_name=action_name,
            action_owner_id=action_owner_id,
            action_digest=action_digest,
            required_inputs=required_inputs,
            for_each=for_each,
            compute_requirement_overrides=compute_requirement_overrides,
            container_parameter_overrides=container_parameter_overrides,
            created_by=created_by,
            condition=condition,
            additional_inputs=additional_inputs,
            parameter_values=parameter_values,
        )
        return cls(
            record=record,
            action_delegate=action_delegate,
            invocation_delegate=invocation_delegate,
            trigger_delegate=trigger_delegate,
        )

    @classmethod
    def from_name(
        cls,
        name: str,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        trigger_delegate: TriggerDelegate,
        org_id: Optional[str] = None,
    ) -> "Trigger":
        record = trigger_delegate.get_trigger_by_primary_key(name=name, org_id=org_id)
        return cls(
            record=record,
            action_delegate=action_delegate,
            invocation_delegate=invocation_delegate,
            trigger_delegate=trigger_delegate,
        )

    @classmethod
    def query(
        cls,
        query: QuerySpecification,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        trigger_delegate: TriggerDelegate,
        org_id: Optional[str] = None,
    ) -> collections.abc.Generator["Trigger", None, None]:
        paginated_results = trigger_delegate.query_triggers(query, org_id=org_id)
        while True:
            for record in paginated_results.items:
                yield cls(
                    record=record,
                    action_delegate=action_delegate,
                    invocation_delegate=invocation_delegate,
                    trigger_delegate=trigger_delegate,
                )
            if paginated_results.next_token:
                query.after = paginated_results.next_token
                paginated_results = trigger_delegate.query_triggers(
                    query, org_id=org_id
                )
            else:
                break

    def __init__(
        self,
        record: TriggerRecord,
        action_delegate: ActionDelegate,
        invocation_delegate: InvocationDelegate,
        trigger_delegate: TriggerDelegate,
    ):
        self.__record = record
        self.__action_delegate = action_delegate
        self.__invocation_delegate = invocation_delegate
        self.__trigger_delegate = trigger_delegate

    @property
    def name(self):
        return self.__record.name

    @property
    def record(self) -> TriggerRecord:
        return self.__record

    @property
    def condition(self) -> Optional[ConditionType]:
        return self.__record.condition

    @property
    def for_each(self) -> TriggerForEachPrimitive:
        return self.__record.for_each

    @property
    def enabled(self) -> bool:
        return self.__record.enabled

    def delete(self):
        self.__trigger_delegate.delete_trigger(
            name=self.__record.name, org_id=self.__record.org_id
        )

    def update(
        self,
        action_name: Union[str, NotSetType] = NotSet,
        action_owner_id: Union[str, NotSetType] = NotSet,
        action_digest: Optional[Union[str, NotSetType]] = NotSet,
        required_inputs: Union[list[str], NotSetType] = NotSet,
        for_each: Union[TriggerForEachPrimitive, NotSetType] = NotSet,
        enabled: Union[bool, NotSetType] = NotSet,
        additional_inputs: Optional[Union[list[str], NotSetType]] = NotSet,
        parameter_values: Optional[Union[dict[str, Any], NotSetType]] = NotSet,
        compute_requirement_overrides: Optional[
            Union[ComputeRequirements, NotSetType]
        ] = NotSet,
        container_parameter_overrides: Optional[
            Union[ContainerParameters, NotSetType]
        ] = NotSet,
        condition: Optional[Union[ConditionType, NotSetType]] = NotSet,
        updated_by: Optional[str] = None,
    ) -> TriggerRecord:
        self.__record = self.__trigger_delegate.update_trigger(
            name=self.__record.name,
            org_id=self.__record.org_id,
            action_name=action_name,
            action_owner_id=action_owner_id,
            action_digest=action_digest,
            required_inputs=required_inputs,
            for_each=for_each,
            enabled=enabled,
            additional_inputs=additional_inputs,
            parameter_values=parameter_values,
            compute_requirement_overrides=compute_requirement_overrides,
            container_parameter_overrides=container_parameter_overrides,
            condition=condition,
            updated_by=updated_by,
        )
        return self.__record

    def action(self) -> Action:
        return Action.from_name(
            name=self.__record.action.name,
            action_delegate=self.__action_delegate,
            invocation_delegate=self.__invocation_delegate,
            org_id=self.__record.org_id,
            digest=self.__record.action.digest,
            action_owner_id=self.__record.action.owner,
        )

    def invoke(
        self,
        data_source: InvocationDataSource,
        idempotency_id: Optional[str] = None,
        input_data_override: Optional[list[str]] = None,
    ) -> Invocation:
        return self.action().invoke(
            input_data=input_data_override or self.__record.required_inputs,
            data_source_id=data_source.data_source_id,
            data_source_type=data_source.data_source_type,
            invocation_source=InvocationSource.Trigger,
            invocation_source_id=self.__record.name,
            parameter_values=self.__record.parameter_values,
            compute_requirement_overrides=self.__record.compute_requirement_overrides,
            container_parameter_overrides=self.__record.container_parameter_overrides,
            idempotency_id=idempotency_id,
            org_id=self.__record.org_id,
        )

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)



#  Copyright (c) 2023 Roboto Technologies, Inc.
import abc
from typing import Optional

from .record import TokenRecord


class TokenDelegate(abc.ABC):
    @abc.abstractmethod
    def get_tokens_for_user(self, user_id: Optional[str] = None) -> list[TokenRecord]:
        raise NotImplementedError("get_tokens_for_user")

    @abc.abstractmethod
    def create_token(
        self,
        expiry_days: int,
        name: str,
        user_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> TokenRecord:
        raise NotImplementedError("create_token")

    @abc.abstractmethod
    def delete_token(self, token_id: str) -> None:
        raise NotImplementedError("delete_token")

    @abc.abstractmethod
    def get_token_by_token_id(self, token_id: str) -> TokenRecord:
        raise NotImplementedError("get_token_by_token_id")



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import json

from ...domain.datasets import Dataset
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    record = Dataset.from_id(
        args.dataset_id, context.datasets, context.files, org_id=args.org
    )
    print(json.dumps(record.to_dict(), indent=4))


def show_setup_parser(parser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )
    add_org_arg(parser)


show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={"help": "Show information about a specific dataset."},
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...serde import pydantic_jsonable_dict
from .delegate import TokenDelegate
from .record import TokenRecord


class Token:
    __record: TokenRecord
    __token_delegate: TokenDelegate

    @classmethod
    def from_id(cls, token_id: str, token_delegate: TokenDelegate) -> "Token":
        record = token_delegate.get_token_by_token_id(token_id=token_id)
        return cls(record=record, token_delegate=token_delegate)

    @classmethod
    def for_user(
        cls, token_delegate: TokenDelegate, user_id: Optional[str] = None
    ) -> list["Token"]:
        records = token_delegate.get_tokens_for_user(user_id=user_id)
        return [cls(record=record, token_delegate=token_delegate) for record in records]

    @classmethod
    def create(
        cls,
        expiry_days: int,
        name: str,
        token_delegate: TokenDelegate,
        user_id: Optional[str] = None,
        description: Optional[str] = None,
    ):
        record = token_delegate.create_token(
            user_id=user_id, expiry_days=expiry_days, name=name, description=description
        )
        return cls(record=record, token_delegate=token_delegate)

    def __init__(self, record: TokenRecord, token_delegate: TokenDelegate):
        self.__record = record
        self.__token_delegate = token_delegate

    @property
    def record(self) -> TokenRecord:
        return self.__record

    @property
    def secret(self) -> Optional[str]:
        return self.__record.secret

    @property
    def user_id(self) -> Optional[str]:
        return self.__record.user_id

    @property
    def token_id(self) -> Optional[str]:
        return None if self.__record.context is None else self.__record.context.token_id

    def delete(self):
        assert self.__record.context is not None
        return self.__token_delegate.delete_token(
            token_id=self.__record.context.token_id
        )

    def to_dict(self, exclude_none: bool = False) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record, exclude_none=exclude_none)



import abc
import enum
import pathlib
from typing import Optional, Protocol

from ...http import PaginatedList
from ...query import QuerySpecification
from .progress import (
    NoopProgressMonitorFactory,
    ProgressMonitorFactory,
)
from .record import FileRecord


class FileTag(enum.Enum):
    DatasetId = "dataset_id"
    OrgId = "org_id"
    # Path to file relative to common prefix
    CommonPrefix = "common_prefix"
    UploadId = "upload_id"


class S3Credentials(Protocol):
    access_key_id: str
    secret_access_key: str
    session_token: str


class FileDelegate(abc.ABC):
    @abc.abstractmethod
    def delete_file(self, record: FileRecord) -> None:
        raise NotImplementedError("delete_file")

    @abc.abstractmethod
    def download_file(
        self,
        record: FileRecord,
        local_path: pathlib.Path,
        credentials: S3Credentials,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ) -> None:
        raise NotImplementedError("download_file")

    @abc.abstractmethod
    def get_record_by_primary_key(
        self, file_id: str, org_id: Optional[str] = None
    ) -> FileRecord:
        raise NotImplementedError("get_record_by_primary_key")

    @abc.abstractmethod
    def get_signed_url(self, record: FileRecord) -> str:
        raise NotImplementedError("get_signed_url")

    @abc.abstractmethod
    def query_files(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[FileRecord]:
        raise NotImplementedError("query_files")

    @abc.abstractmethod
    def upload_file(
        self,
        local_path: pathlib.Path,
        bucket: str,
        key: str,
        credentials: S3Credentials,
        tags: Optional[dict[FileTag, str]] = None,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ) -> None:
        raise NotImplementedError("upload_file")



#  Copyright (c) 2023 Roboto Technologies, Inc.

from .comment import Comment
from .delegate import CommentDelegate
from .http_delegate import CommentHttpDelegate
from .http_resources import (
    CreateCommentRequest,
    UpdateCommentRequest,
)
from .record import CommentRecord, EntityType

__all__ = (
    "CreateCommentRequest",
    "CommentRecord",
    "EntityType",
    "CommentDelegate",
    "Comment",
    "UpdateCommentRequest",
    "CommentHttpDelegate",
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...serde import pydantic_jsonable_dict
from .delegate import OrgDelegate
from .org import Org
from .record import OrgRoleName, OrgRoleRecord


class OrgRole:
    __record: OrgRoleRecord
    __org_delegate: OrgDelegate
    __org: Org

    @classmethod
    def for_user(
        cls, user_id: Optional[str], org_delegate: OrgDelegate
    ) -> list["OrgRole"]:
        records = org_delegate.org_roles_for_user(user_id=user_id)
        return [cls(record=record, org_delegate=org_delegate) for record in records]

    @classmethod
    def for_org(
        cls, org_id: Optional[str], org_delegate: OrgDelegate
    ) -> list["OrgRole"]:
        records = org_delegate.org_roles_for_org(org_id=org_id)
        return [cls(record=record, org_delegate=org_delegate) for record in records]

    @classmethod
    def for_user_in_org(
        cls,
        org_delegate: OrgDelegate,
        user_id: Optional[str] = None,
        org_id: Optional[str] = None,
    ):
        record = org_delegate.org_role_for_user_in_org(user_id=user_id, org_id=org_id)
        return cls(record=record, org_delegate=org_delegate)

    def __init__(self, record: OrgRoleRecord, org_delegate: OrgDelegate):
        self.__record = record
        self.__org_delegate = org_delegate
        self.__org = Org(record=record.org, org_delegate=org_delegate)

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)

    @property
    def org_id(self) -> str:
        return self.__record.org.org_id

    @property
    def org(self) -> Org:
        return self.__org

    @property
    def user_id(self) -> str:
        return self.__record.user.user_id

    @property
    def record(self) -> OrgRoleRecord:
        return self.__record

    @property
    def roles(self) -> list[OrgRoleName]:
        return self.__record.roles



#  Copyright (c) 2023 Roboto Technologies, Inc.

import abc
from typing import Optional

import tqdm


class ProgressMonitor(abc.ABC):
    @abc.abstractmethod
    def update(self, uploaded_bytes: int):
        raise NotImplementedError("update")

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError("close")

    @abc.abstractmethod
    def is_closed(self) -> bool:
        raise NotImplementedError("is_closed")


class ProgressMonitorFactory(abc.ABC):
    @abc.abstractmethod
    def upload_monitor(self, source: str, size: int) -> ProgressMonitor:
        raise NotImplementedError("upload_monitor")

    def download_monitor(self, source: str, size: int) -> ProgressMonitor:
        raise NotImplementedError("download_monitor")


class NoopProgressMonitor(ProgressMonitor):
    def update(self, uploaded_bytes: int):
        pass

    def close(self):
        pass

    def is_closed(self) -> bool:
        return False


class NoopProgressMonitorFactory(ProgressMonitorFactory):
    def upload_monitor(self, source: str, size: int) -> ProgressMonitor:
        return NoopProgressMonitor()

    def download_monitor(self, source: str, size: int) -> ProgressMonitor:
        return NoopProgressMonitor()


class TqdmProgressMonitor(ProgressMonitor):
    __tqdm: tqdm.tqdm
    __is_closed: bool

    def __init__(self, total: int, desc: str, position: int = 0, leave: bool = True):
        self.__tqdm = tqdm.tqdm(
            total=total,
            desc=desc,
            bar_format="{percentage:.1f}%|{bar:25} | {rate_fmt} | {desc}",
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            position=position,
            leave=leave,
        )
        self.__is_closed = False

    def update(self, uploaded_bytes: int):
        self.__tqdm.update(uploaded_bytes)

    def close(self):
        self.__tqdm.close()
        self.__is_closed = True

    def is_closed(self) -> bool:
        return self.__is_closed


class TqdmProgressMonitorFactory(ProgressMonitorFactory):
    __monitors: list[Optional[ProgressMonitor]]

    def __init__(self, concurrency: int = 1):
        self.__monitors = [None] * concurrency

    def __first_available_slot(self) -> Optional[int]:
        for idx in range(len(self.__monitors)):
            if self.__monitors[idx] is None or self.__monitors[idx].is_closed():  # type: ignore[union-attr]
                return idx

        return None

    def __any_monitor(self, source: str, size: int) -> ProgressMonitor:
        # This for sure is not fully threadsafe, but it 100% works for single threading and
        # _mostly_ works for multithreading.
        slot = self.__first_available_slot()
        if slot is None:
            raise ValueError("Number of concurrent monitors is exceeding concurrency!")

        monitor = TqdmProgressMonitor(
            total=size,
            desc=f"Source: {source}",
            position=slot,
            leave=len(self.__monitors) == 1,
        )

        self.__monitors[slot] = monitor

        return monitor

    def upload_monitor(self, source: str, size: int) -> ProgressMonitor:
        return self.__any_monitor(source=source, size=size)

    def download_monitor(self, source: str, size: int) -> ProgressMonitor:
        return self.__any_monitor(source=source, size=size)



import datetime
import enum
import typing
import urllib.parse

import pydantic

from ..auth import Permissions
from ..exceptions import RobotoHttpExceptionParse
from ..http import (
    HttpClient,
    PaginatedList,
    roboto_headers,
)
from ..serde import pydantic_jsonable_dict
from ..time import utcnow
from .http_resources import (
    CreateImageRepositoryRequest,
    DeleteImageRepositoryRequest,
    DeleteImageRequest,
)
from .record import (
    ContainerImageRecord,
    ContainerImageRepositoryRecord,
)


class ContainerCredentials(pydantic.BaseModel):
    username: str
    password: str
    registry_url: str
    expiration: datetime.datetime

    def is_expired(self) -> bool:
        return utcnow() >= self.expiration

    def to_dict(self) -> dict[str, typing.Any]:
        return pydantic_jsonable_dict(self)


class RepositoryPurpose(enum.Enum):
    Executor = "executor"


class RepositoryTag(enum.Enum):
    CreatedBy = "created_by"
    OrgId = "org_id"
    Purpose = "purpose"  # RepositoryPurpose


class ImageRepository(typing.TypedDict):
    repository_name: str
    repository_uri: str


class ImageRegistry:
    __http_client: HttpClient
    __roboto_service_base_url: str

    def __init__(self, roboto_service_base_url: str, http_client: HttpClient) -> None:
        self.__http_client = http_client
        self.__roboto_service_base_url = roboto_service_base_url

    def create_repository(
        self,
        repository_name: str,
        immutable_image_tags: bool = False,
        org_id: typing.Optional[str] = None,
    ) -> ImageRepository:
        """
        Create a repository for a container image in Roboto's private image registry.
        Images with different tags can be pushed to the same repository.

        Args:
            repository_name: The name of the repository to create.
            immutable_image_tags: Whether to allow image tags to be overwritten. If set to True,
                then any attempt to overwrite an existing image tag will error.

        Returns:
            A dictionary contains the `repository_name` and `repository_uri` of the created repository.
        """
        url = f"{self.__roboto_service_base_url}/v1/images/repository"
        request_body = CreateImageRepositoryRequest(
            repository_name=repository_name,
            immutable_image_tags=immutable_image_tags,
        )

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url,
                data=pydantic_jsonable_dict(
                    request_body, exclude_none=True, exclude_unset=True
                ),
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

        return response.from_json(json_path=["data"])

    def delete_image(self, image_uri: str, org_id: typing.Optional[str] = None) -> None:
        """
        Delete a container image from Roboto's private registry.

        Args:
            image_uri: The full URI of the image to delete.
            org_id: ID of organization owning the provided image.
        """
        url = f"{self.__roboto_service_base_url}/v1/images/image"
        request_body = DeleteImageRequest(
            image_uri=image_uri,
        )

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url,
                data=pydantic_jsonable_dict(request_body),
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

    def delete_repository(
        self,
        repository_name: str,
        org_id: typing.Optional[str] = None,
        force: bool = False,
    ) -> None:
        """
        Delete a container image from Roboto's private registry.

        Args:
            repository_name: The name of the repository to delete.
            org_id: ID of organization owning the provided image.
            force: Delete all images in the repository before deleting the repository if the repository is not empty.
        """
        url = f"{self.__roboto_service_base_url}/v1/images/repository"

        payload = DeleteImageRepositoryRequest(
            repository_name=repository_name,
            force=force,
        )

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url,
                data=pydantic_jsonable_dict(payload),
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

    def get_container_image_record(
        self, org_id: str, image_uri: str
    ) -> ContainerImageRecord:
        url_safe_image_uri = urllib.parse.quote_plus(image_uri)
        url = f"{self.__roboto_service_base_url}/v1/images/image/record/{org_id}/{url_safe_image_uri}"

        with RobotoHttpExceptionParse():
            res = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

        return ContainerImageRecord.parse_obj(res.from_json(["data"]))

    def get_temporary_credentials(
        self,
        repository_uri: str,
        permissions: Permissions = Permissions.ReadOnly,
        org_id: typing.Optional[str] = None,
    ) -> ContainerCredentials:
        query_params = {
            "repository_uri": repository_uri,
            "permissions": permissions.value,
        }
        encoded_qs = urllib.parse.urlencode(query_params)
        url = f"{self.__roboto_service_base_url}/v1/images/credentials?{encoded_qs}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

        return ContainerCredentials.parse_obj(response.from_json(json_path=["data"]))

    def list_images(
        self,
        repository_name: typing.Optional[str] = None,
        page_token: typing.Optional[str] = None,
        org_id: typing.Optional[str] = None,
    ) -> PaginatedList[ContainerImageRecord]:
        url = f"{self.__roboto_service_base_url}/v1/images/image/record/list"
        qs_params = dict()
        if repository_name:
            qs_params["repository_name"] = repository_name

        if page_token:
            qs_params["page_token"] = page_token

        if qs_params:
            encoded_qs = urllib.parse.urlencode(qs_params)
            url = f"{url}?{encoded_qs}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

        response_data = response.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                ContainerImageRecord.parse_obj(item) for item in response_data["items"]
            ],
            next_token=response_data["next_token"],
        )

    def list_repositories(
        self,
        page_token: typing.Optional[str] = None,
        org_id: typing.Optional[str] = None,
    ) -> PaginatedList[ContainerImageRepositoryRecord]:
        url = f"{self.__roboto_service_base_url}/v1/images/repository/record/list"
        if page_token:
            encoded_qs = urllib.parse.urlencode({"page_token": page_token})
            url = f"{url}?{encoded_qs}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

        response_data = response.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                ContainerImageRepositoryRecord.parse_obj(item)
                for item in response_data["items"]
            ],
            next_token=response_data["next_token"],
        )

    def repository_contains_image(
        self,
        repository_name: str,
        image_tag: str,
        org_id: typing.Optional[str] = None,
    ) -> bool:
        urlsafe_repository_name = urllib.parse.quote_plus(repository_name)
        urlsafe_image_tag = urllib.parse.quote_plus(image_tag)
        url = f"{self.__roboto_service_base_url}/v1/images/repository/{urlsafe_repository_name}/contains/{urlsafe_image_tag}"  # noqa: E501

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(
                url,
                headers=roboto_headers(
                    org_id, additional_headers={"Content-Type": "application/json"}
                ),
            )

        contains_image = response.from_json(json_path=["data", "contains_image"])
        return contains_image



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import pathlib

from ...domain.datasets import Dataset
from ..command import (
    ExistingPathlibPath,
    RobotoCommand,
)
from ..common_args import add_org_arg
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def upload_files(args, context: CLIContext, parser: argparse.ArgumentParser):
    path: pathlib.Path = args.path
    if args.exclude is not None and not path.is_dir():
        parser.error(
            "Exclude filters are only supported for directory uploads, not single files."
        )
    if args.key is not None and path.is_dir():
        parser.error(
            "Key overrides are only supported for single file uploads, now directories."
        )

    record = Dataset.from_id(
        args.dataset_id, context.datasets, context.files, org_id=args.org
    )

    if path.is_dir():
        record.upload_directory(directory_path=path, exclude_patterns=args.exclude)
    else:
        key = path.name if args.key is None else args.key
        record.upload_file(local_file_path=path, key=key)


def upload_files_setup_parser(parser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )
    parser.add_argument(
        "-k",
        "--key",
        type=str,
        help="A key to alias a file to when storing it to a dataset. Does nothing for directories.",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=ExistingPathlibPath,
        required=True,
        help="The path to a file or directory to upload.",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        nargs="*",
        help="Zero or more exclude filters (if path points to a directory)",
    )
    add_org_arg(parser)


upload_files_command = RobotoCommand(
    name="upload-files",
    logic=upload_files,
    setup_parser=upload_files_setup_parser,
    command_kwargs={"help": "Uploads a file or directory to a specific dataset."},
)



import collections.abc
import hashlib
import pathlib
from typing import Any, Optional

from ...query import QuerySpecification
from ...serde import pydantic_jsonable_dict
from .delegate import FileDelegate, S3Credentials
from .progress import (
    NoopProgressMonitorFactory,
    ProgressMonitorFactory,
)
from .record import FileRecord


class File:
    __delegate: FileDelegate
    __record: FileRecord

    @staticmethod
    def compute_file_id(uri: str) -> str:
        """
        Return a fixed size (32 character), unique ID deterministically from an Object's bucket and key.
        Versioned objects will have the same ID.
        """
        return hashlib.blake2b(uri.encode("utf-8"), digest_size=16).hexdigest()

    @staticmethod
    def construct_s3_obj_arn(bucket: str, key: str) -> str:
        return f"arn:aws:s3:::{bucket}/{key}"

    @staticmethod
    def construct_s3_obj_uri(
        bucket: str, key: str, version: Optional[str] = None
    ) -> str:
        base_uri = f"s3://{bucket}/{key}"
        if version:
            base_uri += f"?versionId={version}"
        return base_uri

    @classmethod
    def from_id(
        cls,
        file_id: str,
        delegate: FileDelegate,
        org_id: Optional[str] = None,
    ) -> "File":
        record = delegate.get_record_by_primary_key(file_id, org_id)
        return cls(record, delegate)

    @classmethod
    def query(
        cls,
        query: QuerySpecification,
        delegate: FileDelegate,
        org_id: Optional[str] = None,
    ) -> collections.abc.Generator["File", None, None]:
        known = set(FileRecord.__fields__.keys())
        actual = set()
        for field in query.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of File"
                if plural
                else "is not a known attribute of File"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        paginated_results = delegate.query_files(query, org_id=org_id)
        while True:
            for record in paginated_results.items:
                yield cls(record, delegate)
            if paginated_results.next_token:
                query.after = paginated_results.next_token
                paginated_results = delegate.query_files(query, org_id=org_id)
            else:
                break

    def __init__(self, record: FileRecord, delegate: FileDelegate):
        self.__record = record
        self.__delegate = delegate

    @property
    def file_id(self) -> str:
        return self.__record.file_id

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def uri(self) -> str:
        return self.__record.uri

    @property
    def record(self) -> FileRecord:
        return self.__record

    @property
    def relative_path(self) -> str:
        return self.__record.relative_path

    def delete(self) -> None:
        self.__delegate.delete_file(self.__record)

    def download(
        self,
        local_path: pathlib.Path,
        credentials: S3Credentials,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ):
        self.__delegate.download_file(
            self.__record,
            local_path,
            credentials,
            progress_monitor_factory=progress_monitor_factory,
        )

    def get_signed_url(self) -> str:
        return self.__delegate.get_signed_url(self.__record)

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)



import abc
from typing import Any, Optional, Union

from ...http import PaginatedList
from ...query import QuerySpecification
from ...sentinels import NotSet, NotSetType
from ...updates import MetadataChangeset
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionRecord,
    ActionReference,
)


class ActionDelegate(abc.ABC):
    @abc.abstractmethod
    def create_action(
        self,
        name: str,
        parameters: Optional[list[ActionParameter]] = None,
        uri: Optional[str] = None,
        inherits: Optional[ActionReference] = None,
        compute_requirements: Optional[ComputeRequirements] = None,
        container_parameters: Optional[ContainerParameters] = None,
        description: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        tags: Optional[list[str]] = None,
        created_by: Optional[str] = None,  # A Roboto user_id
        org_id: Optional[str] = None,
    ) -> ActionRecord:
        raise NotImplementedError("create_action")

    @abc.abstractmethod
    def get_action_by_primary_key(
        self,
        name: str,
        org_id: Optional[str] = None,
        digest: Optional[str] = None,
        action_owner_id: Optional[str] = None,
    ) -> ActionRecord:
        raise NotImplementedError("get_action_by_primary_key")

    @abc.abstractmethod
    def delete_action(self, record: ActionRecord) -> None:
        raise NotImplementedError("delete_action")

    @abc.abstractmethod
    def query_actions(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[ActionRecord]:
        raise NotImplementedError("query_actions")

    @abc.abstractmethod
    def query_actions_on_action_hub(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[ActionRecord]:
        raise NotImplementedError("query_actions_on_action_hub")

    @abc.abstractmethod
    def set_accessibility(
        self,
        record: ActionRecord,
        accessibility: Accessibility,
    ) -> ActionRecord:
        raise NotImplementedError("set_accessibility")

    @abc.abstractmethod
    def update(
        self,
        record: ActionRecord,
        compute_requirements: Union[ComputeRequirements, NotSetType] = NotSet,
        container_parameters: Union[ContainerParameters, NotSetType] = NotSet,
        description: Optional[Union[str, NotSetType]] = NotSet,
        metadata_changeset: Union[MetadataChangeset, NotSetType] = NotSet,
        parameter_changeset: Union[ActionParameterChangeset, NotSetType] = NotSet,
        uri: Union[str, NotSetType] = NotSet,
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> ActionRecord:
        raise NotImplementedError("update")



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...exceptions import RobotoHttpExceptionParse
from ...http import HttpClient, roboto_headers
from ...serde import pydantic_jsonable_dict
from .delegate import OrgDelegate
from .http_resources import (
    BindEmailDomainRequest,
    CreateOrgRequest,
    InviteUserRequest,
    ModifyRoleForUserRequest,
    OrgRecordUpdates,
    RemoveUserFromOrgRequest,
    UpdateOrgRequest,
)
from .record import (
    OrgInviteRecord,
    OrgRecord,
    OrgRoleName,
    OrgRoleRecord,
)


class OrgHttpDelegate(OrgDelegate):
    __http_client: HttpClient

    def __init__(self, http_client: HttpClient):
        super().__init__()
        self.__http_client = http_client

    def create_org(
        self,
        creator_user_id: Optional[str],
        name: str,
        bind_email_domain: bool = False,
    ) -> OrgRecord:
        url = self.__http_client.url("v1/orgs")
        headers = roboto_headers(user_id=creator_user_id)
        headers["Content-Type"] = "application/json"

        request_body = CreateOrgRequest(name=name, bind_email_domain=bind_email_domain)

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

        return OrgRecord.parse_obj(response.from_json(json_path=["data"]))

    def update_org(
        self,
        updates: dict[str, Any],
        org_id: Optional[str] = None,
        caller_user_id: Optional[str] = None,
    ) -> OrgRecord:
        url = self.__http_client.url("v1/orgs")
        headers = roboto_headers(user_id=caller_user_id, org_id=org_id)
        headers["Content-Type"] = "application/json"
        request_body = UpdateOrgRequest(updates=OrgRecordUpdates.parse_obj(updates))

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url=url,
                headers=headers,
                data=pydantic_jsonable_dict(request_body, exclude_unset=True),
            )

        return OrgRecord.parse_obj(response.from_json(json_path=["data"]))

    def orgs_for_user(self, user_id: Optional[str]) -> list[OrgRecord]:
        url = self.__http_client.url("v1/users/orgs")

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)

        return [
            OrgRecord.parse_obj(record)
            for record in response.from_json(json_path=["data"])
        ]

    def org_roles_for_user(self, user_id: Optional[str]) -> list[OrgRoleRecord]:
        url = self.__http_client.url("v1/users/roles")

        headers = roboto_headers(user_id=user_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)

        return [
            OrgRoleRecord.parse_obj(record)
            for record in response.from_json(json_path=["data"])
        ]

    def add_role_for_user(
        self, user_id: str, role_name: OrgRoleName, org_id: Optional[str] = None
    ):
        url = self.__http_client.url("v1/orgs/roles")
        headers = roboto_headers(org_id=org_id)
        request_body = ModifyRoleForUserRequest(user_id=user_id, role_name=role_name)

        with RobotoHttpExceptionParse():
            self.__http_client.put(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

    def remove_role_from_user(
        self, user_id: str, role_name: OrgRoleName, org_id: Optional[str] = None
    ):
        url = self.__http_client.url("v1/orgs/roles")
        headers = roboto_headers(org_id=org_id)
        request_body = ModifyRoleForUserRequest(user_id=user_id, role_name=role_name)

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

    def org_roles_for_org(self, org_id: Optional[str]) -> list[OrgRoleRecord]:
        url = self.__http_client.url("v1/orgs/roles")

        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)

        return [
            OrgRoleRecord.parse_obj(record)
            for record in response.from_json(json_path=["data"])
        ]

    def remove_user_from_org(self, user_id: str, org_id: Optional[str] = None) -> None:
        url = self.__http_client.url("v1/orgs/users")

        headers = roboto_headers(org_id=org_id)

        request_body = RemoveUserFromOrgRequest(user_id=user_id)

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

    def get_org_by_id(self, org_id: str) -> OrgRecord:
        url = self.__http_client.url("v1/orgs")
        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)

        return OrgRecord.parse_obj(response.from_json(json_path=["data"]))

    def delete_org(self, org_id: str) -> None:
        url = self.__http_client.url("v1/orgs")
        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            self.__http_client.delete(url=url, headers=headers)

    def bind_email_domain(self, org_id: str, email_domain: str):
        url = self.__http_client.url("v1/orgs/subdomains")
        headers = roboto_headers(org_id=org_id)
        request_body = BindEmailDomainRequest(email_domain=email_domain)

        with RobotoHttpExceptionParse():
            self.__http_client.put(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

    def unbind_email_domain(self, email_domain: str, org_id: Optional[str] = None):
        url = self.__http_client.url("v1/orgs/subdomains")
        headers = roboto_headers(org_id=org_id)
        request_body = BindEmailDomainRequest(email_domain=email_domain)

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

    def get_email_domains_for_org(self, org_id: Optional[str] = None) -> list[str]:
        url = self.__http_client.url("v1/orgs/subdomains")
        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)
            return response.from_json()["data"]

    def invite_user_to_org(
        self, invited_user_id: str, org_id: str, inviting_user_id: Optional[str] = None
    ) -> OrgInviteRecord:
        url = self.__http_client.url("v1/orgs/invites")
        headers = roboto_headers(org_id=org_id)
        request_body = InviteUserRequest(invited_user_id=invited_user_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url=url, headers=headers, data=pydantic_jsonable_dict(request_body)
            )

        return OrgInviteRecord.parse_obj(response.from_json(json_path=["data"]))

    def accept_org_invite(self, invite_id: str, user_id: Optional[str] = None):
        url = self.__http_client.url(f"v1/orgs/invites/{invite_id}/accept")

        with RobotoHttpExceptionParse():
            self.__http_client.post(url=url)

    def decline_org_invite(self, invite_id: str, user_id: Optional[str] = None):
        url = self.__http_client.url(f"v1/orgs/invites/{invite_id}/decline")

        with RobotoHttpExceptionParse():
            self.__http_client.post(url=url)

    def get_org_invite(
        self, invite_id: str, user_id: Optional[str] = None
    ) -> OrgInviteRecord:
        url = self.__http_client.url(f"v1/orgs/invites/{invite_id}")

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)
        return OrgInviteRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_invites_for_org(
        self, org_id: Optional[str] = None
    ) -> list[OrgInviteRecord]:
        url = self.__http_client.url("v1/orgs/invites")
        headers = roboto_headers(org_id=org_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)

        return [
            OrgInviteRecord.parse_obj(record)
            for record in response.from_json(json_path=["data"])
        ]

    def org_role_for_user_in_org(
        self, user_id: Optional[str] = None, org_id: Optional[str] = None
    ) -> OrgRoleRecord:
        assert user_id is not None

        url = self.__http_client.url(f"v1/orgs/roles/{user_id}")

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)
        return OrgRoleRecord.parse_obj(response.from_json(json_path=["data"]))



import typing

import pydantic


class FileRecordRequest(pydantic.BaseModel):
    """Upsert a file record."""

    bucket: str
    key: str
    version: typing.Optional[str] = None


class DeleteFileRequest(pydantic.BaseModel):
    uri: str


class QueryFilesRequest(pydantic.BaseModel):
    filters: dict[str, typing.Any] = pydantic.Field(default_factory=dict)

    class Config:
        extra = "forbid"


class SignedUrlResponse(pydantic.BaseModel):
    url: str



#  Copyright (c) 2023 Roboto Technologies, Inc.

from .http_delegate import TriggerHttpDelegate
from .http_resources import (
    CreateTriggerRequest,
    EvaluateTriggerPrincipalType,
    EvaluateTriggerScope,
    EvaluateTriggersRequest,
    QueryTriggersRequest,
    UpdateTriggerRequest,
)
from .trigger import Trigger
from .trigger_delegate import TriggerDelegate
from .trigger_record import (
    TriggerForEachPrimitive,
    TriggerRecord,
)

__all__ = [
    "CreateTriggerRequest",
    "EvaluateTriggerPrincipalType",
    "EvaluateTriggerScope",
    "EvaluateTriggersRequest",
    "QueryTriggersRequest",
    "Trigger",
    "TriggerDelegate",
    "TriggerForEachPrimitive",
    "TriggerRecord",
    "TriggerHttpDelegate",
    "UpdateTriggerRequest",
]



#  Copyright (c) 2023 Roboto Technologies, Inc.
from .actions import (
    ActionParameterArg,
    ActionReferenceParser,
    add_action_reference_arg,
    add_compute_requirements_args,
    add_container_parameters_args,
    parse_compute_requirements,
    parse_container_overrides,
)
from .exceptions import ParseError
from .orgs import add_org_arg

__all__ = (
    "ActionParameterArg",
    "ActionReferenceParser",
    "add_action_reference_arg",
    "add_compute_requirements_args",
    "add_container_parameters_args",
    "add_org_arg",
    "parse_compute_requirements",
    "parse_container_overrides",
    "ParseError",
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

import argparse
from importlib.metadata import version
import logging
import pathlib
import sys
from typing import Any, Optional

from ..domain.actions import (
    ActionHttpDelegate,
    InvocationHttpDelegate,
)
from ..domain.datasets import DatasetHttpDelegate
from ..domain.files import FileClientDelegate
from ..domain.orgs import OrgHttpDelegate
from ..domain.tokens import TokenHttpDelegate
from ..domain.triggers import TriggerHttpDelegate
from ..domain.users import UserHttpDelegate
from ..http import (
    HttpClient,
    LocalAuthDecorator,
    PATAuthDecorator,
)
from ..profile import (
    MalformedProfileException,
    MissingProfileException,
    RobotoProfile,
)
from .actions import (
    command_set as actions_command_set,
)
from .argparse import SortingHelpFormatter
from .config import check_last_update
from .context import CLIContext
from .datasets import (
    command_set as datasets_command_set,
)
from .extension import (
    apply_roboto_cli_command_extensions,
    apply_roboto_cli_context_extensions,
)
from .images import (
    command_set as images_command_set,
)
from .invocations import (
    command_set as invocations_command_set,
)
from .orgs import command_set as orgs_command_set
from .tokens import (
    command_set as tokens_command_set,
)
from .triggers import (
    command_set as triggers_command_set,
)
from .users import (
    command_set as users_command_set,
)

COMMAND_SETS = [
    actions_command_set,
    datasets_command_set,
    images_command_set,
    invocations_command_set,
    orgs_command_set,
    users_command_set,
    tokens_command_set,
    triggers_command_set,
]

BETA_USER_POOL_CLIENT_ID = "7p2e45lijin58tuaairtflf3m8"
PROD_USER_POOL_CLIENT_ID = "1gricmdmh0vv582qdd84phab5"


PROGRAMMATIC_ACCESS_BLURB = (
    "To resolve this, please consult the getting started page for programmatic access at "
    + "https://docs.roboto.ai/getting-started/programmatic-access.html."
)


def __populate_context(profile: RobotoProfile, context: CLIContext):
    profile_entry = profile.get_entry()

    http_client_args: dict[str, Any] = {
        "default_endpoint": profile_entry.default_endpoint
    }

    if "localhost" not in profile_entry.default_endpoint:
        http_client_args["default_auth"] = PATAuthDecorator.for_client(
            client_id=profile_entry.default_client_id, profile=profile
        )
    else:
        http_client_args["default_auth"] = LocalAuthDecorator(
            user_id=profile_entry.user_id
        )

    http = HttpClient(**http_client_args)  # type: ignore[arg-type]
    context.roboto_service_base_url = profile_entry.default_endpoint
    context.http = http

    context.actions = ActionHttpDelegate(
        http_client=http, roboto_service_base_url=profile_entry.default_endpoint
    )
    context.datasets = DatasetHttpDelegate(
        http_client=http, roboto_service_base_url=profile_entry.default_endpoint
    )
    context.files = FileClientDelegate(
        http_client=http, roboto_service_base_url=profile_entry.default_endpoint
    )
    context.invocations = InvocationHttpDelegate(
        http_client=http, roboto_service_base_url=profile_entry.default_endpoint
    )
    context.orgs = OrgHttpDelegate(http_client=http)
    context.triggers = TriggerHttpDelegate(http_client=http)
    context.users = UserHttpDelegate(http_client=http)
    context.tokens = TokenHttpDelegate(http_client=http)
    context.extensions = {}


def construct_parser(context: Optional[CLIContext] = None) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="roboto",
        description=(
            "CLI for interacting with Roboto's Data Platform. "
            "Each of the command groups listed below have their own set of supported subcommands and help pages."
        ),
        formatter_class=SortingHelpFormatter,
    )

    parser.add_argument(
        "--debug",
        help="Set the log level to DEBUG, which will print verbose logs to STDERR.",
        action="store_true",
    )

    parser.add_argument(
        "--version",
        help="Show the version of 'roboto_sdk' currently running",
        action="store_true",
    )

    parser.add_argument(
        "--profile",
        help="Roboto profile to use; must match a section within the Roboto config.json",
        required=False,
    )

    parser.add_argument(
        "--config-file",
        help="Overrides the location of the roboto config.json file. Defaults to ~/.roboto/config.json",
        type=pathlib.Path,
        required=False,
    )

    # https://bugs.python.org/issue29298
    subcommands = parser.add_subparsers(dest="function")

    for command_set in sorted(
        apply_roboto_cli_command_extensions(base_command_sets=COMMAND_SETS),
        key=lambda x: x.name,
    ):
        command_set.sort_commands()
        command_set.add_to_subparsers(subcommands, context)

    return parser


def entry():
    logging.basicConfig(level=logging.ERROR, stream=sys.stderr)

    context = CLIContext()
    parser = construct_parser(context)
    profile = RobotoProfile()

    try:
        # By default, as soon as we figure out we're running a sub-parser, any field encountered after that subparser
        # will be ignored. This means that `roboto --debug datasets search` will work but
        # `roboto datasets search --debug` will not. In order to work around this, we can use parser_known_args,
        # which gives us back the list of un-evaluated args, and then take a second pass at parse_args with those args.
        # This will only work if our subparsers never re-define top level parameters like --debug, --config-file, etc.
        #
        # This solution was based on a stack overflow post about this issue:
        # https://stackoverflow.com/questions/46962065/add-top-level-argparse-arguments-after-subparser-args
        args, unparsed = parser.parse_known_args()
        args = parser.parse_args(unparsed, args)

        profile = (
            RobotoProfile(default_profile_name=args.profile)
            if args.config_file is None
            else RobotoProfile(
                config_file=args.config_file, default_profile_name=args.profile
            )
        )

        try:
            profile.get_entry()
        except MalformedProfileException:
            parser.error(
                f"The roboto config file located at '{profile.config_file}' is malformed. "
                + PROGRAMMATIC_ACCESS_BLURB
            )
        except MissingProfileException:
            parser.error(
                f"No roboto config file located at '{profile.config_file}'. "
                + PROGRAMMATIC_ACCESS_BLURB
            )

        __populate_context(profile, context)
        apply_roboto_cli_context_extensions(base_context=context)

        if args.debug:
            logging.getLogger().setLevel(logging.DEBUG)
        else:
            logging.getLogger().setLevel(logging.ERROR)

        if args.version:
            print(version("roboto_sdk"))
        elif "func" in args:
            args.func(args)
        else:
            parser.print_help()
    finally:
        check_last_update(profile)
